# MESMER, land-climate dynamics group, S.I. Seneviratne
# Copyright (c) 2021 ETH Zurich, MESMER contributors listed in AUTHORS.
# Licensed under the GNU General Public License v3.0 or later see LICENSE or
# https://www.gnu.org/licenses/

import joblib
import xarray as xr


def save_mesmer_bundle(
    bundle_file,
    params_lt,
    params_lv,
    params_gv,
    land_fractions,
    lat,
    lon,
):
    """
    Save all the information required to draw MESMER emulations to disk

    Parameters
    ----------
    bundle_file : str
        file in which to save the bundle
    params_lt : dict
        dictionary containing the calibrated parameters for the local trends emulations,
        keys relevant here:

        - ["targs"] (list of emulated variables, list of strs)
        - ["esm"] (Earth System Model, str)
        - ["method"] (applied method, str)
        - [xx] (additional keys depending on employed method)
    params_lv : dict
        dictionary containing the calibrated parameters for the local variability
        emulations, keys relevant here

        - ["targs"] (list of variables which are emulated, list of strs)
        - ["esm"] (Earth System Model, str)
        - ["method"] (applied method, str)
        - [xx] (additional keys depending on employed method)
    params_gv : dict
        dictionary containing the calibrated parameters for the global variability emulations, keys relevant here

        - ["targ"] (variable which is emulated, str)
        - ["esm"] (Earth System Model, str)
        - ["method"] (applied method, str)
        - ["preds"] (predictors, list of strs)
        - ["scenarios"] (scenarios which are used for training, list of strs)
        - [xx] (additional keys depending on employed method)
    land_fractions : np.MaskedArray
        data containing land fractions (also used for helping generate output on lat-lon
        grids)
    lat : np.ndarray
        grid latitudes (used to check land_fractions shape)
    lon : np.ndarray
        grid longitudes (used to check land_fractions shape)
    """
    assert land_fractions.shape[0] == lat.shape[0]
    assert land_fractions.shape[1] == lon.shape[0]

    # hopefully right way around
    land_fractions = xr.DataArray(
        land_fractions, dims=["lat", "lon"], coords={"lat": lat, "lon": lon}
    )

    mesmer_bundle = {
        "params_lt": params_lt,
        "params_lv": params_lv,
        "params_gv": params_gv,
        "land_fractions": land_fractions,
    }
    joblib.dump(mesmer_bundle, bundle_file)



# code vendored from regionmask under the conditions of their license
# see licenses/REGIONMASK_LICENSE

import numpy as np
import xarray as xr


def mask_percentage(regions, lon, lat):
    """Sample with 10 times higher resolution.

    Notes
    -----
    - assumes equally-spaced lat & lon!
    - copied from Mathias Hauser: https://github.com/mathause/regionmask/issues/38 in
      August 2020
    - prototype of what will eventually be integrated in his regionmask package

    """

    lon_sampled = sample_coord(lon)
    lat_sampled = sample_coord(lat)

    mask = regions.mask(lon_sampled, lat_sampled)

    isnan = np.isnan(mask.values)

    numbers = np.unique(mask.values[~isnan])
    numbers = numbers.astype(np.int)

    mask_sampled = list()
    for num in numbers:
        # coarsen the mask again
        mask_coarse = (mask == num).coarsen(lat=10, lon=10).mean()
        mask_sampled.append(mask_coarse)

    mask_sampled = xr.concat(
        mask_sampled, dim="region", compat="override", coords="minimal"
    )

    mask_sampled = mask_sampled.assign_coords(region=("region", numbers))

    return mask_sampled


def sample_coord(coord):
    """Sample coords for the percentage overlap.

    Notes
    -----
    - copied from Mathias Hauser: https://github.com/mathause/regionmask/issues/38
      in August 2020
    -> prototype of what will eventually be integrated in his regionmask package

    """
    d_coord = coord[1] - coord[0]

    n_cells = len(coord)

    left = coord[0] - d_coord / 2 + d_coord / 20
    right = coord[-1] + d_coord / 2 - d_coord / 20

    return np.linspace(left, right, n_cells * 10)



# code vendored from xarray under the conditions of their license
# see licenses/XARRAY_LICENSE

import numpy as np

# TODO: potentially use xr.plot.utils._infer_interval_breaks. However, that's not
# public API so certainly need tests first.


def infer_interval_breaks(x, y, clip=False):
    """Find edges of gridcells, given their centers."""

    if len(x.shape) == 1:
        x = _infer_interval_breaks(x)
        y = _infer_interval_breaks(y)
    else:
        # we have to infer the intervals on both axes
        x = _infer_interval_breaks(x, axis=1)
        x = _infer_interval_breaks(x, axis=0)
        y = _infer_interval_breaks(y, axis=1)
        y = _infer_interval_breaks(y, axis=0)

    if clip:
        y = np.clip(y, -90, 90)

    return x, y


def _infer_interval_breaks(coord, axis=0):
    """
    >>> _infer_interval_breaks(np.arange(5))
    array([-0.5,  0.5,  1.5,  2.5,  3.5,  4.5])
    >>> _infer_interval_breaks([[0, 1], [3, 4]], axis=1)
    array([[-0.5,  0.5,  1.5],
           [ 2.5,  3.5,  4.5]])
    """

    if not _is_monotonic(coord, axis=axis):
        raise ValueError(
            "The input coordinate is not sorted in increasing "
            "order along axis %d. This can lead to unexpected "
            "results. Consider calling the `sortby` method on "
            "the input DataArray. To plot data with categorical "
            "axes, consider using the `heatmap` function from "
            "the `seaborn` statistical plotting library." % axis
        )

    coord = np.asarray(coord)
    deltas = 0.5 * np.diff(coord, axis=axis)
    if deltas.size == 0:
        deltas = np.array(0.0)
    first = np.take(coord, [0], axis=axis) - np.take(deltas, [0], axis=axis)
    last = np.take(coord, [-1], axis=axis) + np.take(deltas, [-1], axis=axis)
    trim_last = tuple(
        slice(None, -1) if n == axis else slice(None) for n in range(coord.ndim)
    )
    return np.concatenate([first, coord[trim_last] + deltas, last], axis=axis)


def _is_monotonic(coord, axis=0):
    """
    >>> _is_monotonic(np.array([0, 1, 2]))
    True
    >>> _is_monotonic(np.array([2, 1, 0]))
    True
    >>> _is_monotonic(np.array([0, 2, 1]))
    False
    """
    coord = np.asarray(coord)

    if coord.shape[axis] < 3:
        return True
    else:
        n = coord.shape[axis]
        delta_pos = coord.take(np.arange(1, n), axis=axis) >= coord.take(
            np.arange(0, n - 1), axis=axis
        )
        delta_neg = coord.take(np.arange(1, n), axis=axis) <= coord.take(
            np.arange(0, n - 1), axis=axis
        )

    return np.all(delta_pos) or np.all(delta_neg)



# MESMER, land-climate dynamics group, S.I. Seneviratne
# Copyright (c) 2021 ETH Zurich, MESMER contributors listed in AUTHORS.
# Licensed under the GNU General Public License v3.0 or later see LICENSE or
# https://www.gnu.org/licenses/
"""
Functions to create local trends emulations with MESMER.
"""


import os

import joblib
import numpy as np


def create_emus_lt(params_lt, preds_lt, cfg, concat_h_f=False, save_emus=True):
    """
    Create local trends (i.e., forced response) emulations for given parameter set and
    predictors.

    Parameters
    ----------
    params_lt : dict
        dictionary with the trained local trend parameters

        - ["targs"] (emulated variables, str)
        - ["esm"] (Earth System Model, str)
        - ["method"] (applied method, str)
        - ["method_each_gp_sep"] (states if method is applied to each grid point
          separately, bool)
        - ["preds"] (predictors, list of strs)
        - ["scenarios"] (emission scenarios used for training, list of strs)
        - [xx] (additional params depend on method employed)
        - ["full_model_contains_lv"] (whether the full model contains part of the local
          variability module, bool)
    preds_lt : dict
        nested dictionary of predictors for local trends with keys

        - [pred][scen] (1d/2d arrays (time)/(run, time) of predictor for specific scenario)
    cfg : module
        config file containing metadata
    concat_h_f : bool, optional
        determines if historical and future time period is concatenated into a single
        emulation or not, default = False (must be set to False if no historical data
        provided)
    save_emus : bool, optional
        determines if parameters are saved or not, default = True

    Returns
    -------
    emus_lt : dict
        local trend emulations nested dictionary with keys

        - [scen]["targ"] (2d array (time, gp) of local trend emulations)

    Notes
    -----
    - Assumptions:
        - same predictors for each target
        - if historical time period is included in predictors, it has its own dictionary
          key
        - if historical time period was included in training, it has its own scenario
        - either historical period is included for every scenario or for no scenario
    - Potential TODO:
        - evaluate if really need / want concat_h_f or if I want output to be determined
          by shape predictors

    """
    # specify necessary variables from config file
    if save_emus:
        dir_mesmer_emus = cfg.dir_mesmer_emus

    # derive necessary scenario names
    pred_names = list(preds_lt.keys())
    scenarios_emus = list(preds_lt[pred_names[0]].keys())

    if concat_h_f:
        if scenarios_emus[0] == "hist":
            scens_out_f = scenarios_emus[1:]
            scens_out = ["h-" + s for s in scens_out_f]
        else:
            raise ValueError("This combination is not supported.")
    else:
        if "h-" in scenarios_emus[0]:
            scens_out_f = list(map(lambda x: x.replace("h-", ""), scenarios_emus))
            scens_out = ["hist"] + scens_out_f
        else:
            scens_out = scens_out_f = scenarios_emus

    # check predictors
    if pred_names != params_lt["preds"]:  # check if correct predictors
        raise ValueError(
            "Wrong predictors were passed. The emulations cannot be created."
        )

    # select the method from a dict of fcts
    create_emus_method_func_mapping = {"OLS_each_gp_sep": create_emus_OLS_each_gp_sep}
    # extend dict if add more methods

    # dict could be extended to contain other methods (even actual fcts that I write)
    method_lt = params_lt["method"]
    if params_lt["method_each_gp_sep"]:
        method_lt = method_lt + "_each_gp_sep"
    else:
        raise ValueError(
            f"No such method ({params_lt['method_each_gp_sep']}) is currently implemented."
        )

    create_emus_method_lt = create_emus_method_func_mapping[method_lt]

    # create emulations
    emus_lt = {}
    if concat_h_f:
        lt_hist = create_emus_method_lt(params_lt, preds_lt, "hist")
        for scen_out, scen_out_f in zip(scens_out, scens_out_f):
            lt_scen_f = create_emus_method_lt(params_lt, preds_lt, scen_out_f)
            emus_lt[scen_out] = {}
            for targ in params_lt["targs"]:
                emus_lt[scen_out][targ] = np.concatenate(
                    [lt_hist[targ], lt_scen_f[targ]]
                )
    else:
        for scen_out in scens_out:
            emus_lt[scen_out] = create_emus_method_lt(params_lt, preds_lt, scen_out)

    # save the local trends emulation if requested
    if save_emus:
        dir_mesmer_emus_lt = dir_mesmer_emus + "local/local_trends/"
        # check if folder to save params in exists, if not: make it
        if not os.path.exists(dir_mesmer_emus_lt):
            os.makedirs(dir_mesmer_emus_lt)
            print("created dir:", dir_mesmer_emus_lt)
        filename_parts = [
            "emus_lt",
            params_lt["method"],
            *params_lt["preds"],
            *params_lt["targs"],
            params_lt["esm"],
            *scens_out,
        ]
        filename_emus_lt = dir_mesmer_emus_lt + "_".join(filename_parts) + ".pkl"
        joblib.dump(emus_lt, filename_emus_lt)

    return emus_lt


def create_emus_OLS_each_gp_sep(params_lt, preds_lt, scen):
    """Create local trends with OLS with grid-point-specific predictors

    Parameters
    ----------
    params_lt : dict
        dictionary with the trained local trend parameters

        - ["targs"] (emulated variables, str)
        - ["esm"] (Earth System Model, str)
        - ["method"] (applied method, str)
        - ["method_each_gp_sep"] (states if method is applied to each grid point
          separately, bool)
        - ["preds"] (predictors, list of strs)
        - ["scenarios"] (emission scenarios used for training, list of strs)
        - [xx] (additional params depend on method employed)
        - ["full_model_contains_lv"] (whether the full model contains part of the local
          variability module, bool)
    preds_lt : dict
        nested dictionary of predictors for local trends with keys

        - [pred][scen] (1d/ 2d arrays (time)/(run, time) of predictor for specific scenario)
    scen : str
        emulated scenario

    Returns
    -------
    emus_lt : dict
        local trend emulations dictionary with keys

        - ["targ"] (2d array (time, gp) of local trend emulations)

    Notes
    -----
    - Assumptions:
        - Coefficients are the same for every scenario

    """

    pred_names = list(preds_lt.keys())
    nr_ts = len(
        preds_lt[pred_names[0]][scen]
    )  # nr_ts could vary for different scenarios but is the same for all predictors

    emus_lt = {}
    for targ in params_lt["targs"]:
        nr_gps = len(params_lt["intercept"][targ])
        emus_lt[targ] = np.zeros([nr_ts, nr_gps])
        for gp in np.arange(nr_gps):
            pred_vals = [
                params_lt["coef_" + pred][targ][gp] * preds_lt[pred][scen]
                for pred in params_lt["preds"]
            ]

            emus_lt[targ][:, gp] = sum(pred_vals) + params_lt["intercept"][targ][gp]

    return emus_lt



# MESMER, land-climate dynamics group, S.I. Seneviratne
# Copyright (c) 2021 ETH Zurich, MESMER contributors listed in AUTHORS.
# Licensed under the GNU General Public License v3.0 or later see LICENSE or
# https://www.gnu.org/licenses/
"""
Functions to process data.
"""


import numpy as np


def convert_dict_to_arr(var_dict):
    """Convert dictionary to array.

    Parameters
    ----------
    var_dict : dict
        nested variable (e.g., tas) dictionary with keys

        - [scen][run] (xd array (time, x) of variable)

    Returns
    -------
    var_arr : dict
        variable dictionary with keys

        - [scen] (xd array (run, time, x) of variable)

    """

    scenarios = list(var_dict.keys())

    var_arr = {}

    for scen in scenarios:
        runs = list(var_dict[scen])
        shape_run = list(var_dict[scen][runs[0]].shape)
        var_arr[scen] = np.zeros([len(runs)] + shape_run)

        for i in np.arange(len(runs)):
            var_arr[scen][i] = var_dict[scen][runs[i]]

    return var_arr


def separate_hist_future(var_c, time_c, cfg):
    """Separate historical and future time periods into separate keys in dictionary.

    Parameters
    ----------
    var_c : dict
        variable dictionary with concatenated historical and future scenarios as keys

        - ["h-scen_f"] (xd array of variable (run, time, x), np.ndarray)
    time_c : dict
        time dictionary with concatenated historical and future scenarios as keys

        - ["h-scen_f"] (1d array of years, np.ndarray)
    cfg : module
        config file containing metadata

    Returns
    -------
    var_s : dict
        variable dictionary with separated historical and future scenarios as keys

        - [hist] / [scen_f] (xd array of variable (run, time, x), np.ndarray)
    time_s : dict
        time dictionary with separated historical and future scenarios as keys

        - ["hist] / [scen_f] (1d array of years, np.ndarray)

    Notes
    -----
    - Assumption
        - each scenario starts in the same year
        - at least 2-dim variable array is passed with 1st dim nr runs, 2nd dim nr ts

    """

    gen = cfg.gen
    scens_c = list(var_c.keys())  # concatenated scens
    scens_f = list(map(lambda x: x.replace("h-", ""), scens_c))  # future scens

    if gen == 5:
        end_year_hist = 2005  # last year included in historical time period
    if gen == 6:
        end_year_hist = 2014

    scen_c = scens_c[0]
    scen_f = scens_f[0]
    idx_start_fut = np.where(time_c[scen_c] == end_year_hist)[0][0] + 1

    time_s = {}
    time_s["hist"] = time_c[scen_c][:idx_start_fut]

    var_s = {}
    var_s["hist"] = var_c[scen_c][:, :idx_start_fut]
    var_s[scen_f] = var_c[scen_c][:, idx_start_fut:]
    time_s[scen_f] = time_c[scen_c][idx_start_fut:]
    for scen_f, scen_c in zip(scens_f[1:], scens_c[1:]):
        var_s["hist"] = np.vstack(
            [var_s["hist"], var_c[scen_c][:, :idx_start_fut]]
        )  # stack all available historical runs
        var_s[scen_f] = var_c[scen_c][:, idx_start_fut:]
        time_s[scen_f] = time_c[scen_c][idx_start_fut:]

    # exclude duplicate historical runs that are available in several scenarios
    var_s["hist"] = np.unique(var_s["hist"], axis=0)

    return var_s, time_s



# MESMER, land-climate dynamics group, S.I. Seneviratne
# Copyright (c) 2021 ETH Zurich, MESMER contributors listed in AUTHORS.
# Licensed under the GNU General Public License v3.0 or later see LICENSE or
# https://www.gnu.org/licenses/
"""
Functions to create local variability emulations with MESMER.
"""


import os

import joblib
import numpy as np


def create_emus_lv(params_lv, preds_lv, cfg, save_emus=True, submethod=""):
    """Create local variablity emulations.

    Parameters
    ----------
    params_lv : dict
        dictionary with the trained local variability parameters

        - ["targ"] (variable which is emulated, str)
        - ["esm"] (Earth System Model, str)
        - ["method"] (applied method, str)
        - ["preds"] (predictors, list of strs)
        - ["scenarios"] (scenarios which are used for training, list of strs)
        - [xx] (additional keys depend on employed method)
    preds_lv : dict
        nested dictionary of predictors for local variability with keys

        - [pred][scen] (1d/ 2d arrays (time)/(run, time) of predictor for specific scenario)
    cfg : module
        config file containing metadata
    save_emus : bool, optional
        determines if emulation is saved or not, default = True
    submethod : str, optional
        determines if only submethod should be used, default = "" indicating using the
        full method

    Returns
    -------
    emus_lv : dict
        local variability emulations dictionary with keys

        - [scen] (3d array  (emu, time, gp) of local variability emulation time series)

    Notes
    -----
    - Assumptions:
        - if stochastic realizations are drawn, preds_lv must contain concatenated
          hist + future scenarios or only future scenarios
        - if no preds_lv needed, pass time as predictor instead such that can get info
          about how many scenarios / ts per scenario should be drawn for stochastic part
        - submethod specific assumptions are listed in the submethod description
    - Long-term TODO:
        - improve consistency with actual esms by sharing same realizations in
          historical time period and diff one for each scen in future
        - improve this function in terms of generalization properties + design
          (+ consistency with rest of code)
        - improve function to also work if only part of predictors are used for OLS and
          others for other methods
    """

    # specify necessary variables from config file
    if save_emus:
        dir_mesmer_emus = cfg.dir_mesmer_emus

    pred_names = list(preds_lv.keys())
    scens_out = list(preds_lv[pred_names[0]].keys())

    # if no submethod is specified, the actual method contained in params_lv is employed
    if len(submethod) == 0:
        submethod = params_lv["method"]

    # carry out emulations
    emus_lv = {}

    if "OLS" in submethod and params_lv["method_lt_each_gp_sep"]:
        emus_lv = create_emus_lv_OLS(params_lv, preds_lv)
        # overwrites empty emus_lv dict, i.e., must always be executed as first method

    # HERE ALTERNATIVE METHODS COULD BE PLACED WHICH ARE EXECUTED BEFORE AR1_sci
    # e.g., my idea to do hybrid between Link et al and my things

    if "AR1_sci" in submethod:
        emus_lv = create_emus_lv_AR1_sci(emus_lv, params_lv, preds_lv, cfg)

    # save the local trends emulation if requested
    if save_emus:
        dir_mesmer_emus_lv = dir_mesmer_emus + "local/local_variability/"
        # check if folder to save params in exists, if not: make it
        if not os.path.exists(dir_mesmer_emus_lv):
            os.makedirs(dir_mesmer_emus_lv)
            print("created dir:", dir_mesmer_emus_lv)
        filename_parts = [
            "emus_lv",
            submethod,
            *params_lv["preds"],
            *params_lv["targs"],
            params_lv["esm"],
            *scens_out,
        ]
        filename_emus_lv = dir_mesmer_emus_lv + "_".join(filename_parts) + ".pkl"
        joblib.dump(emus_lv, filename_emus_lv)

    return emus_lv


def create_emus_lv_AR1_sci(emus_lv, params_lv, preds_lv, cfg):
    """
    Create local variablity emulations with AR(1) process with spatially-correlated
    innovations.

    Parameters
    ----------
    emus_lv : dict
        local variability emulations dictionary with keys

        - [scen] (3d array (emu, time, gp) of local variability from previous submethods)
        - empty dict if no previous submethod
    params_lv : dict
        dictionary with the trained local variability parameters

        - ["targ"] (variable which is emulated, str)
        - ["esm"] (Earth System Model, str)
        - ["method"] (applied method, str)
        - ["preds"] (predictors, list of strs)
        - ["scenarios"] (scenarios which are used for training, list of strs)
        - [xx] (additional keys depend on employed method)
    preds_lv : dict
        nested dictionary of predictors for local variability with keys

        - [pred][scen] (1d/ 2d arrays (time)/(run, time) of predictor for specific scenario)
    cfg : module
        config file containing metadata

    Returns
    -------
    emus_lv : dict
        local variability emulations dictionary with keys

        - [scen] (2d array  (emu, time, gp) of local variability in response to global
          variability emulation time series)

    Notes
    -----
    - Assumptions:
        - do for each target variable independently.
        - the variability is Gaussian
    - Long-term TODO:
        - add possibility to account for cross-correlation between different variables

    """

    print("Start with AR(1) with spatially correlated innovations.")
    pred_names = list(preds_lv.keys())
    scens_out = list(preds_lv[pred_names[0]].keys())
    nr_emus_v = cfg.nr_emus_v
    seed_all_scens = cfg.seed[params_lv["esm"]]

    for scen in scens_out:
        # if 1-d array, time = 1st dim, else time = 2nd dim
        if len(preds_lv[pred_names[0]][scen].shape) > 1:
            nr_ts_emus_stoch_v = preds_lv[pred_names[0]][scen].shape[1]
        else:
            nr_ts_emus_stoch_v = preds_lv[pred_names[0]][scen].shape[0]

        if scen not in emus_lv:
            emus_lv[scen] = {}

        for targ in params_lv["targs"]:

            seed = seed_all_scens[scen]["lv"]
            nr_gps = len(params_lv["AR1_int"][targ])

            # in case no emus_lv[scen] exist yet, initialize it. Otherwise build up on existing one
            if len(emus_lv[scen]) == 0:

                emus_lv[scen][targ] = np.zeros(nr_emus_v, nr_ts_emus_stoch_v, nr_gps)

            # ensure reproducibility
            np.random.seed(seed)

            # buffer so that initial start at 0 does not influence overall result
            buffer = 20

            print("Draw the innovations")
            # draw the innovations
            innovs = np.random.multivariate_normal(
                np.zeros(nr_gps),
                params_lv["loc_ecov_AR1_innovs"][targ],
                size=[nr_emus_v, nr_ts_emus_stoch_v + buffer],
            )

            print(
                "Compute the contribution to emus_lv by the AR(1) process with the spatially correlated innovations"
            )
            emus_lv_tmp = np.zeros([nr_emus_v, nr_ts_emus_stoch_v + buffer, nr_gps])
            for t in np.arange(1, nr_ts_emus_stoch_v + buffer):
                emus_lv_tmp[:, t, :] = (
                    params_lv["AR1_int"][targ]
                    + params_lv["AR1_coef"][targ] * emus_lv_tmp[:, t - 1, :]
                    + innovs[:, t, :]
                )
            emus_lv_tmp = emus_lv_tmp[:, buffer:, :]

            print("Create the full local variability emulations")
            emus_lv[scen][targ] += emus_lv_tmp

    return emus_lv


def create_emus_lv_OLS(params_lv, preds_lv):
    """Create local variablity emulations with OLS.

    Parameters
    ----------
    params_lv : dict
        dictionary with the trained local variability parameters

        - ["targ"] (variable which is emulated, str)
        - ["esm"] (Earth System Model, str)
        - ["method"] (applied method, str)
        - ["preds"] (predictors, list of strs)
        - ["scenarios"] (scenarios which are used for training, list of strs)
        - [xx] (additional keys depend on employed method)
    preds_lv : dict
        nested dictionary of predictors for local variability with keys

        - [pred][scen] (1d/ 2d arrays (time)/(run, time) of predictor for specific scenario)

    Returns
    -------
    emus_lv : dict
        local variability emulations dictionary with keys

        - [scen] (3d array  (emu, time, gp) of local variability in response to global
          variability emulation time series)

    Notes
    -----
    - Assumptions:
        - first submethod that gets executed (i.e., assumes can make a new emus_lv dict
          within this function)
        - all predictors in preds_lv are being used (ie no other part of method is
          allowed to have predictors)
        - OLS coefs are the same for each scenario
    """

    print("Start with OLS")
    pred_names = list(preds_lv.keys())
    if pred_names != params_lv["preds"]:
        raise ValueError("Wrong list of predictors was passed.")

    scens_OLS = list(preds_lv[pred_names[0]].keys())
    emus_lv = {}
    for scen in scens_OLS:
        emus_lv[scen] = {}

        for targ in params_lv["targs"]:
            nr_emus_v, nr_ts_emus_v = preds_lv[pred_names[0]][scen].shape
            nr_gps = len(params_lv["coef_" + params_lv["preds"][0]][targ])
            emus_lv[scen][targ] = np.zeros([nr_emus_v, nr_ts_emus_v, nr_gps])
            for run in np.arange(nr_emus_v):
                for gp in np.arange(nr_gps):
                    emus_lv[scen][targ][run, :, gp] = sum(
                        [
                            params_lv["coef_" + pred][targ][gp]
                            * preds_lv[pred][scen][run]
                            for pred in params_lv["preds"]
                        ]
                    )
    return emus_lv



# MESMER, land-climate dynamics group, S.I. Seneviratne
# Copyright (c) 2021 ETH Zurich, MESMER contributors listed in AUTHORS.
# Licensed under the GNU General Public License v3.0 or later see LICENSE or
# https://www.gnu.org/licenses/
"""
Functions to merge emulations of different MESMER modules.
"""


import os
import warnings

import joblib


def create_emus_g(emus_gt, emus_gv, params_gt, params_gv, cfg, save_emus=True):
    """Merge global trend and global variability emulations of the same scenarios.

    Parameters
    ----------
    emus_gt : dict
        global trend emulations dictionary with keys

        - [scen] (1d array of global trend emulation time series)
    emus_gv : dict
        global variability emulations dictionary with keys

        - [scen] (2d array  (emus, time) of global variability emulation time series)
    params_gt : dict
        dictionary containing the calibrated parameters for the global trend emulations,
        keys relevant here:

        - ["targ"] (emulated variable, str)
        - ["esm"] (Earth System Model, str)
        - ["method"] (applied method, str)
    params_gv : dict
        dictionary containing the calibrated parameters for the global variability
        emulations, keys relevant here:

        - ["targ"] (variable which is emulated, str)
        - ["esm"] (Earth System Model, str)
        - ["method"] (applied method, str)
    cfg : module
        config file containing metadata
    save_emus : bool, optional
        determines if emulation is saved or not, default = True

    Returns
    -------
    emus_g : dict
        global emulations dictionary with keys

        - [scen] (2d array (emus, time) of global emulation time series)
    """
    scenarios_gt = list(emus_gt.keys())
    scenarios_gv = list(emus_gv.keys())

    emus_g = {}
    if scenarios_gt == scenarios_gv:
        for scen in scenarios_gt:
            emus_g[scen] = emus_gt[scen] + emus_gv[scen]
    elif scenarios_gv == ["all"]:
        for scen in scenarios_gt:
            emus_g[scen] = emus_gt[scen] + emus_gv["all"]
    else:
        warnings.warn(
            "The global trend and the global variabilty emulations are not from the"
            " same scenario, no global emulation is created."
        )
        emus_g = []

    if params_gt["targ"] == params_gv["targ"]:
        targ = params_gt["targ"]
    else:
        warnings.warn(
            "The target variables do not match. No global emulation is created."
        )
        emus_g = []

    if params_gt["esm"] == params_gv["esm"]:
        esm = params_gt["esm"]
    else:
        warnings.warn(
            "The Earth System Models do not match. No global emulation is created."
        )
        emus_g = []

    # save the global emus if requested
    if save_emus:
        dir_mesmer_emus = cfg.dir_mesmer_emus
        dir_mesmer_emus_g = dir_mesmer_emus + "global/"
        # check if folder to save emus in exists, if not: make it
        if not os.path.exists(dir_mesmer_emus_g):
            os.makedirs(dir_mesmer_emus_g)
            print("created dir:", dir_mesmer_emus_g)

        filename_parts = [
            "emus_g",
            "gt",
            params_gt["method"],
            *params_gt["preds"],
            "gv",
            params_gv["method"],
            *params_gv["preds"],
            targ,
            esm,
            *scenarios_gt,
        ]
        filename_emus_g = dir_mesmer_emus_g + "_".join(filename_parts) + ".pkl"
        joblib.dump(emus_g, filename_emus_g)

    return emus_g


def create_emus_l(emus_lt, emus_lv, params_lt, params_lv, cfg, save_emus=True):
    """
    Merge local trends and local variability temperature emulations of the same
    scenarios and targets.

    Parameters
    ----------
    emus_lt : dict
        local trend emulations dictionary with keys

        - [scen][targ] (2d array (time x grid points) of local trends emulation time
          series)
    emus_lv : dict
        local variability emulations dictionary with keys

        - [scen][targ] (3d array  (emus x time x grid points) of local varaibility
          emulation time series)
    params_lt : dict
        dictionary containing the calibrated parameters for the local trends emulations,
        keys relevant here

        - ["targs"] (list of emulated variables, list of strs)
        - ["esm"] (Earth System Model, str)
        - ["method"] (applied method, str)
    params_lv : dict
        dictionary containing the calibrated parameters for the local variability
        emulations, keys relevant here

        - ["targs"] (list of variables which are emulated, list of strs)
        - ["esm"] (Earth System Model, str)
        - ["method"] (applied method, str)
    cfg : module
        config file containing metadata
    save_emus : bool, optional
        determines if emulation is saved or not, default = True

    Returns
    -------
    emus_l : dict
        local emulations dictionary with keys

        - [scen][targ] (3d array  (emus, time, grid points) of local emulation time
          series)

    """

    # specify necessary variables from config file
    if save_emus:
        dir_mesmer_emus = cfg.dir_mesmer_emus

    scenarios_lt = list(emus_lt.keys())
    scenarios_lv = list(emus_lv.keys())

    targs_lt = list(emus_lt[scenarios_lt[0]].keys())
    targs_lv = list(emus_lv[scenarios_lv[0]].keys())

    emus_l = {}
    if scenarios_lt == scenarios_lv and targs_lt == targs_lv:
        for scen in scenarios_lt:
            emus_l[scen] = {}
            for targ in targs_lt:
                emus_l[scen][targ] = emus_lt[scen][targ] + emus_lv[scen][targ]
    elif scenarios_lv == ["all"] and targs_lt == targs_lv:
        for scen in scenarios_lt:
            emus_l[scen] = {}
            for targ in targs_lt:
                emus_l[scen][targ] = emus_lt[scen][targ] + emus_lv["all"][targ]
    else:
        warnings.warn("The scenarios do not match. No local emulation is created.")
        emus_l = []

    if targs_lt == targs_lv:
        targs = params_lt["targs"]
    else:
        warnings.warn(
            "The target variables do not match. No local emulation is created."
        )
        emus_l = []

    if params_lt["esm"] == params_lv["esm"]:
        esm = params_lt["esm"]
    else:
        warnings.warn(
            "The Earth System Models do not match. No local emulation is created."
        )
        emus_l = []

    # save the global emus if requested
    if save_emus:
        dir_mesmer_emus_l = dir_mesmer_emus + "local/"
        # check if folder to save emus in exists, if not: make it
        if not os.path.exists(dir_mesmer_emus_l):
            os.makedirs(dir_mesmer_emus_l)
            print("created dir:", dir_mesmer_emus_l)
        filename_parts = [
            "emus_l",
            "lt",
            params_lt["method"],
            *params_lt["preds"],
            "lv",
            params_lv["method"],
            *params_lv["preds"],
            *targs,
            esm,
            *scenarios_lt,
        ]
        filename_emus_l = dir_mesmer_emus_l + "_".join(filename_parts) + ".pkl"
        joblib.dump(emus_l, filename_emus_l)

    return emus_l



# MESMER, land-climate dynamics group, S.I. Seneviratne
# Copyright (c) 2021 ETH Zurich, MESMER contributors listed in AUTHORS.
# Licensed under the GNU General Public License v3.0 or later see LICENSE or
# https://www.gnu.org/licenses/
"""
Functions to train global variability module of MESMER.
"""


import os

import joblib
import numpy as np
import statsmodels.api as sm
from statsmodels.tsa.ar_model import AutoReg, ar_select_order


def train_gv(gv, targ, esm, cfg, save_params=True, **kwargs):
    """
    Derive global variability parameters for a specified method.

    Parameters
    ----------
    gv : dict
        Nested global mean variability dictionary with keys

        - [scen] (2d array (run, time) of globally-averaged variability time series)
    targ : str
        target variable (e.g., "tas")
    esm : str
        associated Earth System Model (e.g., "CanESM2" or "CanESM5")
    cfg : config module
        config file containing metadata
    save_params : bool, optional
        determines if parameters are saved or not, default = True
    **kwargs:
        additional arguments, passed through to the training function

    Returns
    -------
    params_gv : dict
        dictionary containing the trained parameters for the chosen method / ensemble
        type

        - ["targ"] (emulated variable, str)
        - ["esm"] (Earth System Model, str)
        - ["method"] (applied method, str)
        - ["preds"] (predictors, list of strs)
        - ["scenarios"] (emission scenarios used for training, list of strs)
        - [xx] additional params depend on method employed, specified in
          ``train_gv_T_method()`` function

    Notes
    -----
    - Assumption
        - if historical data is used for training, it has its own scenario

    - TODO:
        - add ability to weight samples differently than equal weight for each scenario

    """

    # specify necessary variables from config file
    method_gv = cfg.methods[targ]["gv"]
    preds_gv = cfg.preds[targ]["gv"]
    wgt_scen_tr_eq = cfg.wgt_scen_tr_eq

    scenarios_tr = list(gv.keys())

    # initialize parameters dictionary and fill in the metadata which does not depend on
    # the applied method
    params_gv = {}
    params_gv["targ"] = targ
    params_gv["esm"] = esm
    params_gv["method"] = method_gv
    params_gv["preds"] = preds_gv
    params_gv["scenarios"] = scenarios_tr

    # apply the chosen method
    if params_gv["method"] == "AR" and wgt_scen_tr_eq:
        # specifiy parameters employed for AR process fitting
        if "max_lag" not in kwargs:
            kwargs["max_lag"] = 12
        if "sel_crit" not in kwargs:
            kwargs["sel_crit"] = "bic"
        params_gv = train_gv_AR(params_gv, gv, kwargs["max_lag"], kwargs["sel_crit"])
    else:
        raise ValueError(
            "The chosen method and / or weighting approach is currently not implemented."
        )

    # save the global variability paramters if requested
    if save_params:
        dir_mesmer_params = cfg.dir_mesmer_params
        dir_mesmer_params_gv = dir_mesmer_params + "global/global_variability/"
        # check if folder to save params in exists, if not: make it
        if not os.path.exists(dir_mesmer_params_gv):
            os.makedirs(dir_mesmer_params_gv)
            print("created dir:", dir_mesmer_params_gv)
        filename_parts = [
            "params_gv",
            method_gv,
            *preds_gv,
            targ,
            esm,
            *scenarios_tr,
        ]
        filename_params_gv = dir_mesmer_params_gv + "_".join(filename_parts) + ".pkl"
        joblib.dump(params_gv, filename_params_gv)

    return params_gv


def train_gv_AR(params_gv, gv, max_lag, sel_crit):
    """
    Derive AR parameters of global variability under the assumption that gv does not
    depend on the scenario.

    Parameters
    ----------
    params_gv : dict
        parameter dictionary containing keys which do not depend on applied method

        - ["targ"] (variable, i.e., tas or tblend, str)
        - ["esm"] (Earth System Model, str)
        - ["method"] (applied method, i.e., AR, str)
        - ["scenarios"] (emission scenarios used for training, list of strs)
    gv : dict
        nested global mean temperature variability (volcanic influence removed)
        dictionary with keys

        - [scen] (2d array (nr_runs, nr_ts) of globally-averaged temperature variability
          time series)
    max_lag: int
        maximum number of lags considered during fitting
    sel_crit: str
        selection criterion for the AR process order, e.g., 'bic' or 'aic'

    Returns
    -------
    params : dict
        parameter dictionary containing original keys plus

        - ["max_lag"] (maximum lag considered when finding suitable AR model, hardcoded
          to 15 here, int)
        - ["sel_crit"] (selection criterion applied to find suitable AR model, hardcoded
          to Bayesian Information Criterion bic here, str)
        - ["AR_int"] (intercept of the AR model, float)
        - ["AR_coefs"] (coefficients of the AR model for the lags which are contained in
          the selected AR model, list of floats)
        - ["AR_order_sel"] (selected AR order, int)
        - ["AR_std_innovs"] (standard deviation of the innovations of the selected AR
          model, float)

    Notes
    -----
    - Assumptions
        - number of runs per scenario and the number of time steps in each scenario can
        vary
        - each scenario receives equal weight during training

    """

    params_gv["max_lag"] = max_lag
    params_gv["sel_crit"] = sel_crit

    # select the AR Order
    nr_scens = len(gv.keys())
    AR_order_scens_tmp = np.zeros(nr_scens)

    for scen_idx, scen in enumerate(gv.keys()):
        nr_runs = gv[scen].shape[0]
        AR_order_runs_tmp = np.zeros(nr_runs)

        for run in np.arange(nr_runs):
            run_ar_lags = ar_select_order(
                gv[scen][run], maxlag=max_lag, ic=sel_crit, old_names=False
            ).ar_lags
            # if order > 0 is selected,add selected order to vector
            if len(run_ar_lags) > 0:
                AR_order_runs_tmp[run] = run_ar_lags[-1]

        AR_order_scens_tmp[scen_idx] = np.percentile(
            AR_order_runs_tmp, q=50, interpolation="nearest"
        )
        # interpolation is not a good way to go here because it could lead to an AR
        # order that wasn't chosen by run -> avoid it by just taking nearest

    AR_order_sel = int(np.percentile(AR_order_scens_tmp, q=50, interpolation="nearest"))

    # determine the AR params for the selected AR order
    params_gv["AR_int"] = 0
    params_gv["AR_coefs"] = np.zeros(AR_order_sel)
    params_gv["AR_order_sel"] = AR_order_sel
    params_gv["AR_std_innovs"] = 0

    for scen_idx, scen in enumerate(gv.keys()):
        nr_runs = gv[scen].shape[0]
        AR_order_runs_tmp = np.zeros(nr_runs)
        AR_int_tmp = 0
        AR_coefs_tmp = np.zeros(AR_order_sel)
        AR_std_innovs_tmp = 0

        for run in np.arange(nr_runs):
            AR_model_tmp = AutoReg(
                gv[scen][run], lags=AR_order_sel, old_names=False
            ).fit()
            AR_int_tmp += AR_model_tmp.params[0] / nr_runs
            AR_coefs_tmp += AR_model_tmp.params[1:] / nr_runs
            AR_std_innovs_tmp += np.sqrt(AR_model_tmp.sigma2) / nr_runs

        params_gv["AR_int"] += AR_int_tmp / nr_scens
        params_gv["AR_coefs"] += AR_coefs_tmp / nr_scens
        params_gv["AR_std_innovs"] += AR_std_innovs_tmp / nr_scens

    # check if fitted AR process is stationary
    # (highly unlikely this test will ever fail but better safe than sorry)
    ar = np.r_[1, -params_gv["AR_coefs"]]  # add zero-lag and negate
    ma = np.r_[1]  # add zero-lag
    arma_process = sm.tsa.ArmaProcess(ar, ma)
    if not arma_process.isstationary:
        raise ValueError(
            "The fitted AR process is not stationary. Another solution is needed."
        )

    return params_gv



# MESMER, land-climate dynamics group, S.I. Seneviratne
# Copyright (c) 2021 ETH Zurich, MESMER contributors listed in AUTHORS.
# Licensed under the GNU General Public License v3.0 or later see LICENSE or
# https://www.gnu.org/licenses/
"""
Functions to train local variability module of MESMER.
"""


import os

import joblib
import numpy as np
from scipy.stats import multivariate_normal
from statsmodels.tsa.ar_model import AutoReg

from .train_utils import train_l_prepare_X_y_wgteq


def train_lv(preds, targs, esm, cfg, save_params=True, aux={}, params_lv={}):
    """Derive local variability (i.e., natural variabiliy) parameters.

    Parameters
    ----------
    preds : dict
        empty dictionary if none, else nested dictionary of predictors with keys

        - [pred][scen]  (1d/ 2d arrays (time)/(run, time) of predictor for specific
        scenario)
    targs : dict
        nested dictionary of targets with keys

        - [targ][scen] (3d array (run, time, gp) of target for specific scenario)
    esm : str
        associated Earth System Model (e.g., "CanESM2" or "CanESM5")
    cfg : module
        config file containing metadata
    save_params : bool, optional
        determines if parameters are saved or not, default = True
    aux : dict, optional
        provides auxiliary variables needed for lv method at hand

        - [var] (Xd arrays of auxiliary variable)
    params_lv : dict, optional
        pass the params_lv dict, if it already exists so that builds upon that one

    Returns
    -------
    params_lv : dict
        dictionary of local variability paramters

        - ["targs"] (emulated variables, str)
        - ["esm"] (Earth System Model, str)
        - ["method"] (applied method, str)
        - ["preds"] (predictors, list of strs)
        - ["scenarios"] (emission scenarios used for training, list of strs)
        - ["part_model_in_lt"] (states if part of the model is saved in params_lt, bool)
        - ["method_lt_each_gp_sep"] (states if local trends method is applied to each
          grid point separately, bool)
        - [xx] (additional params depend on employed lv method)

    Notes
    -----
    - Assumptions:
        - all targets use same approach and same predictors
        - each predictor and each target has the same scenarios as keys
        - all available scenarios are used for training
        - identified parameters are valid for all training scenarios
        - if historical data is used for training, it has its own scenario
        - need to pass the params_lv dict if it already exists so that can continue to
          build on it
    - Disclaimer:
        - currently no method with preds implemented; but already have in there for
          consistency
    - TODO:
        - add ability to weight samples differently than equal weight for each scenario
        in AR process

    """

    targ_names = list(targs.keys())
    targ_name = targ_names[0]  # because same approach for each targ
    pred_names = list(preds.keys())

    # specify necessary variables from config file
    wgt_scen_tr_eq = cfg.wgt_scen_tr_eq

    preds_lv = []
    # check if any preds from pr
    if len(params_lv) > 0:
        [preds_lv.append(pred) for pred in params_lv["preds"]]
    # for now only gv implemented, but could easily extend to rv (regional) lv (local)
    # if wanted such preds
    for pred in pred_names:
        if "gv" in pred:
            preds_lv.append(pred)
    # add new predictors to params_lv
    if len(params_lv) > 0:
        params_lv["preds"] = preds_lv

    method_lv = cfg.methods[targ_name]["lv"]

    scenarios_tr = list(targs[targ_name].keys())

    # prepare predictors and targets
    X, y, wgt_scen_eq = train_l_prepare_X_y_wgteq(preds, targs)
    if wgt_scen_tr_eq is False:
        wgt_scen_eq[:] = 1  # each sample same weight

    if len(params_lv) == 0:
        print("Initialize params_lv dictionary")
        params_lv = {}
        params_lv["targs"] = targ_names
        params_lv["esm"] = esm
        params_lv["method"] = method_lv
        params_lv["preds"] = preds_lv
        params_lv["scenarios"] = scenarios_tr
        params_lv["part_model_in_lt"] = False

    if "AR1_sci" in method_lv and wgt_scen_tr_eq:

        # assumption: target values I feed in here is already ready for AR1_sci method
        # if were to add any other method before (ie introduce Link et al method for
        # large-scale teleconnections), would have to execute it first & fit this one on
        # residuals

        params_lv = train_lv_AR1_sci(params_lv, targs, y, wgt_scen_eq, aux, cfg)
    else:
        raise ValueError(
            "The chosen method and / or weighting approach is not implemented."
        )

    # overwrites lv module if already exists, i.e., assumption: always lt before lv
    if save_params:
        dir_mesmer_params = cfg.dir_mesmer_params
        dir_mesmer_params_lv = dir_mesmer_params + "local/local_variability/"
        # check if folder to save params in exists, if not: make it
        if not os.path.exists(dir_mesmer_params_lv):
            os.makedirs(dir_mesmer_params_lv)
            print("created dir:", dir_mesmer_params_lv)
        filename_parts = [
            "params_lv",
            method_lv,
            *preds_lv,
            *targ_names,
            esm,
            *scenarios_tr,
        ]
        filename_params_lv = dir_mesmer_params_lv + "_".join(filename_parts) + ".pkl"
        joblib.dump(params_lv, filename_params_lv)

    return params_lv


def train_lv_AR1_sci(params_lv, targs, y, wgt_scen_eq, aux, cfg):
    """Derive parameters for AR(1) process with spatially-correlated innovations.

    Parameters
    ----------
    params_lv : dict
        dictionary with the trained local variability parameters

        - ["targ"] (variable which is emulated, str)
        - ["esm"] (Earth System Model, str)
        - ["method"] (applied method, str)
        - ["preds"] (predictors, list of strs)
        - ["scenarios"] (scenarios which are used for training, list of strs)
        - [xx] (additional keys depend on employed method)
    targs : dict
        nested dictionary of targets with keys

        - [targ][scen] with 3d arrays (run, time, gp)
    y : np.ndarray
        3d array (sample, gp, targ) of targets
    wgt_scen_eq : np.ndarray
        1d array (sample) of sample weights
    aux : dict
        provides auxiliary variables needed for lv method at hand

        - ["phi_gc"] (Xd arrays of auxiliary variable)
    cfg : module
        config file containing metadata

    Returns
    -------
    emus_lv : dict
        local variability emulations dictionary with keys

        - [scen] (2d array  (emu, time, gp) of local variability in response to global
          variability emulation time series)

    Notes
    -----
    - Assumptions:
        - do for each target variable independently
        - the variability is Gaussian
        - each scenario receives the same weight during training
    - Potential TODO:
        - add possibility to account for cross-correlation between different variables
          (i.e., joint instead of independent emulation)

    """

    print("Derive parameters for AR(1) processes with spatially correlated innovations")
    # AR(1)
    params_lv["AR1_int"] = {}
    params_lv["AR1_coef"] = {}
    params_lv["AR1_std_innovs"] = {}
    params_lv["L"] = {}  # localisation radius
    # empirical cov matrix of the local variability trained on here
    params_lv["ecov"] = {}
    params_lv["loc_ecov"] = {}  # localized empirical cov matrix
    # localized empirical cov matrix of the innovations of the AR(1) process
    params_lv["loc_ecov_AR1_innovs"] = {}

    # largely ignore prepared targets and use original ones instead because in original
    # easier to loop over individ runs / scenarios
    targ_names = list(targs.keys())
    scenarios_tr = list(targs[targ_names[0]].keys())
    nr_scens = len(scenarios_tr)

    # fit parameters for each target individually
    for t, targ_name in enumerate(targ_names):
        targ = targs[targ_name]
        nr_gps = y.shape[1]
        y_targ = y[:, :, t]

        # AR(1)
        params_lv["AR1_int"][targ_name] = np.zeros(nr_gps)
        params_lv["AR1_coef"][targ_name] = np.zeros(nr_gps)
        params_lv["AR1_std_innovs"][targ_name] = np.zeros(nr_gps)

        for scen in scenarios_tr:
            nr_runs, nr_ts, nr_gps = targ[scen].shape
            AR1_int_runs = np.zeros(nr_gps)
            AR1_coef_runs = np.zeros(nr_gps)
            AR1_std_innovs_runs = np.zeros(nr_gps)

            for run in np.arange(nr_runs):
                for gp in np.arange(nr_gps):
                    AR1_model = AutoReg(
                        targ[scen][run, :, gp], lags=1, old_names=False
                    ).fit()
                    AR1_int_runs[gp] += AR1_model.params[0] / nr_runs
                    AR1_coef_runs[gp] += AR1_model.params[1] / nr_runs
                    # sqrt of variance = standard deviation
                    AR1_std_innovs_runs[gp] += np.sqrt(AR1_model.sigma2) / nr_runs

            params_lv["AR1_int"][targ_name] += AR1_int_runs / nr_scens
            params_lv["AR1_coef"][targ_name] += AR1_coef_runs / nr_scens
            params_lv["AR1_std_innovs"][targ_name] += AR1_std_innovs_runs / nr_scens

        # determine localization radius, empirical cov matrix, and localized ecov matrix
        (
            params_lv["L"][targ_name],
            params_lv["ecov"][targ_name],
            params_lv["loc_ecov"][targ_name],
        ) = train_lv_find_localized_ecov(y_targ, wgt_scen_eq, aux, cfg)

        # ATTENTION: STILL NEED TO CHECK IF THIS IS TRUE. I UNFORTUNATELY LEARNED THAT I
        # WROTE THIS FORMULA DIFFERENTLY IN THE ESD PAPER!!!!!!! (But I am pretty sure
        # that code is correct and the error is in the paper)
        # compute localized cov matrix of the innovations of the AR(1) process
        loc_ecov_AR1_innovs = np.zeros(params_lv["loc_ecov"][targ_name].shape)
        for i in np.arange(nr_gps):
            for j in np.arange(nr_gps):
                loc_ecov_AR1_innovs[i, j] = (
                    np.sqrt(1 - params_lv["AR1_coef"][targ_name][i] ** 2)
                    * np.sqrt(1 - params_lv["AR1_coef"][targ_name][j] ** 2)
                    * params_lv["loc_ecov"][targ_name][i, j]
                )

        params_lv["loc_ecov_AR1_innovs"][targ_name] = loc_ecov_AR1_innovs
        # derive the localized ecov of the innovations of the AR(1) process (ie the one
        # I will later draw innovs from)

    return params_lv


def train_lv_find_localized_ecov(y, wgt_scen_eq, aux, cfg):
    """
    Find suitable localization radius for empirical covariance matrix and derive
    localized empirical cov matrix.

    Parameters
    ----------
    y : np.ndarray
        2d array (sample, gp) of specific target
    wgt_scen_eq : np.ndarray
        1d array (sample) of sample weights
    aux : dict
        provides auxiliary variables needed for lv method at hand

        - ["phi_gc"] (dict with localisation radii as keys and each containing a 2d
        array (gp, gp) of of Gaspari-Cohn correlation matrix
    cfg : module
        config file containing metadata

    Returns
    -------
    L_sel : numpy.int64
        selected localization radius
    ecov : np.ndarray
        2d empirical covariance matrix array (gp, gp)
    loc_ecov : np.ndarray
        2d localized empirical covariance matrix array (gp, gp)

    Notes
    -----
    - Function could also handle determining ecov of several variables but would all
      have to be passed in same 2d y array (with corresponding wgt_scen_eq,
      aux["phi_gc"] shapes)

    """

    # derive the indices for the cross validation
    max_iter_cv = cfg.max_iter_cv
    nr_samples = y.shape[0]
    nr_it = np.min([nr_samples, max_iter_cv])
    idx_cv_out = np.zeros([nr_it, nr_samples], dtype=bool)
    for i in np.arange(nr_it):
        idx_cv_out[i, i::max_iter_cv] = True

    # spatial cross-correlations with specified cross val folds
    L_set = np.sort(list(aux["phi_gc"].keys()))  # the Ls to loop through

    llh_max = -10000
    llh_cv_sum = {}
    idx_L = 0
    L_sel = L_set[idx_L]
    idx_break = False

    while (idx_break is False) and (L_sel < L_set[-1]):
        # experience tells: once stop selecting larger loc radii, will not start again
        # better to stop once max is reached (to limit computational effort + amount of
        # singular matrices)
        L = L_set[idx_L]
        llh_cv_sum[L] = 0

        for it in np.arange(nr_it):
            # extract folds
            y_est = y[~idx_cv_out[it]]  # to estimate params
            y_cv = y[idx_cv_out[it]]  # to crossvalidate the estimate
            wgt_scen_eq_est = wgt_scen_eq[~idx_cv_out[it]]
            wgt_scen_eq_cv = wgt_scen_eq[idx_cv_out[it]]

            # compute ecov and likelihood of out fold to be drawn from it
            ecov = np.cov(y_est, rowvar=False, aweights=wgt_scen_eq_est)
            loc_ecov = aux["phi_gc"][L] * ecov
            # we want the mean of the res to be 0
            mean_0 = np.zeros(aux["phi_gc"][L].shape[0])

            llh_cv_each_sample = multivariate_normal.logpdf(
                y_cv, mean=mean_0, cov=loc_ecov, allow_singular=True
            )
            # allow_singular = True because stms ran into singular matrices
            # ESMs eg affected: CanESM2, CanESM5, IPSL-CM5A-LR, MCM-UA-1-0
            # -> reassuring that saw that in these ESMs L values where matrix
            # is not singular yet can end up being selected

            # each cv sample gets its own likelihood -> can sum them up for overall
            # likelihood
            # sum over all samples = wgt average * nr_samples
            llh_cv_fold_sum = np.average(
                llh_cv_each_sample, weights=wgt_scen_eq_cv
            ) * len(wgt_scen_eq_cv)

            # add to full sum over all folds
            llh_cv_sum[L] += llh_cv_fold_sum

        idx_L += 1

        if llh_cv_sum[L] > llh_max:
            L_sel = L
            llh_max = llh_cv_sum[L]
            print("Newly selected L =", L_sel)
        else:
            print("Final selected L =", L_sel)
            idx_break = True

    ecov = np.cov(y, rowvar=False, aweights=wgt_scen_eq)
    loc_ecov = aux["phi_gc"][L_sel] * ecov

    return L_sel, ecov, loc_ecov



# MESMER, land-climate dynamics group, S.I. Seneviratne
# Copyright (c) 2021 ETH Zurich, MESMER contributors listed in AUTHORS.
# Licensed under the GNU General Public License v3.0 or later see LICENSE or
# https://www.gnu.org/licenses/
"""
Collection of functions to calibrate all modules of MESMER.
"""
# flake8: noqa
from .calibrate_mesmer import _calibrate_and_draw_realisations
from .train_gt import *
from .train_gv import *
from .train_lt import *
from .train_lv import *
from .train_utils import *



# MESMER, land-climate dynamics group, S.I. Seneviratne
# Copyright (c) 2021 ETH Zurich, MESMER contributors listed in AUTHORS.
# Licensed under the GNU General Public License v3.0 or later see LICENSE or
# https://www.gnu.org/licenses/
"""
Functions to create global variability emulations with MESMER.
"""


import os

import joblib
import numpy as np


def create_emus_gv(params_gv, preds_gv, cfg, save_emus=True):
    """Create global variablity emulations for specified method.

    Parameters
    ----------
    params_gv : dict
        Parameters dictionary.

        - ["targ"] (variable which is emulated, str)
        - ["esm"] (Earth System Model, str)
        - ["method"] (applied method, str)
        - ["preds"] (predictors, list of strs)
        - ["scenarios"] (scenarios which are used for training, list of strs)
        - [xx] (additional keys depend on employed method and are listed in
          train_gv_T_method() function)
    preds_gv : dict
        nested dictionary of predictors for global variability with keys

        - [pred][scen]  (1d/2d arrays (time)/(run, time) of predictor for specific
        scenario)
    cfg : module
        config file containing metadata
    save_emus : bool, optional
        determines if emulation is saved or not, default = True

    Returns
    -------
    emus_gv : dict
        global variability emulations dictionary with keys

        - [scen] (2d array  (emus, time) of global trend emulation time series)

    Notes
    -----
    - Assumptions:
        - if no preds_gv needed, pass time as predictor instead such that can get info
          about how many scenarios / ts per scenario should be drawn for stochastic part

    """

    # specify necessary variables from config file
    if save_emus:
        dir_mesmer_emus = cfg.dir_mesmer_emus

    nr_emus_v = cfg.nr_emus_v
    seed_all_scens = cfg.seed[params_gv["esm"]]

    pred_names = list(preds_gv.keys())
    scens_out = list(preds_gv[pred_names[0]].keys())

    if scens_out != list(seed_all_scens.keys()):
        raise ValueError(
            "The scenarios which should be emulated do not have a seed assigned in the"
            " config file. The emulations cannot be created."
        )

    # set up dict for emulations of global variability with emulated scenarios as keys
    emus_gv = {}

    for scen in scens_out:
        if len(preds_gv[pred_names[0]][scen].shape) > 1:
            nr_ts_emus_v = preds_gv[pred_names[0]][scen].shape[1]
        else:
            nr_ts_emus_v = preds_gv[pred_names[0]][scen].shape[0]

        # apply the chosen method
        if params_gv["method"] == "AR":
            emus_gv[scen] = create_emus_gv_AR(
                params_gv, nr_emus_v, nr_ts_emus_v, seed_all_scens[scen]["gv"]
            )
        else:
            raise ValueError("The chosen method is currently not implemented.")
            # if the emus should depend on the scen, scen needs to be passed to the fct

    # save the global variability emus if requested
    if save_emus:
        dir_mesmer_emus_gv = dir_mesmer_emus + "global/global_variability/"
        # check if folder to save emus in exists, if not: make it
        if not os.path.exists(dir_mesmer_emus_gv):
            os.makedirs(dir_mesmer_emus_gv)
            print("created dir:", dir_mesmer_emus_gv)
        filename_parts = [
            "emus_gv",
            params_gv["method"],
            *params_gv["preds"],
            params_gv["targ"],
            params_gv["esm"],
            *scens_out,
        ]
        filename_emus_gv = dir_mesmer_emus_gv + "_".join(filename_parts) + ".pkl"
        joblib.dump(emus_gv, filename_emus_gv)

    return emus_gv


def create_emus_gv_AR(params_gv, nr_emus_v, nr_ts_emus_v, seed):
    """Draw global variablity emulations from an AR process.

    Parameters
    ----------
    params_gv : dict
        Parameters dictionary.

        - ["targ"] (variable which is emulated, str)
        - ["esm"] (Earth System Model, str)
        - ["method"] (applied method, str)
        - ["preds"] (predictors, list of strs)
        - ["scenarios"] (scenarios which are used for training, list of strs)
        - ["max_lag"] (maximum lag considered when finding suitable AR model, int)
        - ["sel_crit"] (selection criterion applied to find suitable AR model, str)
        - ["AR_int"] (intercept of the AR model, float)
        - ["AR_coefs"] (coefficients of the AR model for the lags which are contained in
          the selected AR model, list of floats)
        - ["AR_order_sel"] (selected AR order, int)
        - ["AR_std_innovs"] (standard deviation of the innovations of the selected AR
          model, float)
    nr_emus_v : int
        number of global variability emulations
    nr_ts_emus_v : int
        number of time steps in each global variability emulation
    seed : int
        esm and scenario specific seed for gv module to ensure reproducability of
        results

    Returns
    -------
    emus_gv : dict
        global variability emulations dictionary with keys

        - [scen] (2d array  (emus, time) of global variability emulation time series)
    """

    # ensure reproducibility
    np.random.seed(seed)

    # buffer so that initial start at 0 does not influence overall result
    buffer = 50

    # re-name params for easier reading of code below
    ar_int = params_gv["AR_int"]
    ar_coefs = params_gv["AR_coefs"]
    ar_lags = np.arange(1, params_gv["AR_order_sel"] + 1, dtype=int)

    # if AR(0) process chosen, no AR_coefs are available -> to have code run
    # nevertheless ar_coefs and ar_lags are set to 0 (-> emus are created with
    # ar_int + innovs)
    if len(ar_coefs) == 0:
        ar_coefs = [0]
        ar_lags = [0]

    innovs_emus_gv = np.random.normal(
        loc=0, scale=params_gv["AR_std_innovs"], size=(nr_emus_v, nr_ts_emus_v + buffer)
    )

    # initialize global variability emulations (dim array: nr_emus x nr_ts)
    emus_gv = np.zeros([nr_emus_v, nr_ts_emus_v + buffer])

    for i in np.arange(nr_emus_v):
        # simulate from AR process
        for t in np.arange(ar_lags[-1], len(emus_gv[i])):  # avoid misleading indices
            emus_gv[i, t] = (
                ar_int
                + sum(
                    [
                        ar_coefs[k] * emus_gv[i, t - ar_lags[k]]
                        for k in np.arange(len(ar_lags))
                    ]
                )
                + innovs_emus_gv[i, t]
            )
    # remove buffer
    emus_gv = emus_gv[:, buffer:]

    return emus_gv



# MESMER, land-climate dynamics group, S.I. Seneviratne
# Copyright (c) 2021 ETH Zurich, MESMER contributors listed in AUTHORS.
# Licensed under the GNU General Public License v3.0 or later see LICENSE or
# https://www.gnu.org/licenses/
"""
Functions to extract regions or time period of interest.
"""


import copy as copy

import numpy as np


def extract_land(var, reg_dict, wgt, ls, threshold_land=0.25):
    """
    Extract all land grid points and area weights in regions and in land-sea mask for
    given threshold.

    Parameters
    ----------
    var : dict
        nested variable dictionary with keys

        - [esm][scen] (4d array (run, time, lat, lon) of variable)
    reg_dict : dict
        region dictionary with keys

        - ["type"] (region type)
        - ["abbrevs"] (abbreviations for regions)
        - ["names"] (full names of regions)
        - ["grids"] (3d array (regions, lat, lon) of subsampled region fraction)
        - ["grid_b"] (2d array (lat, lon) of regions with each grid point being assigned
          to a single region ("binary" grid))
        - ["full"] (full Region object (for plotting region outlines))
    wgt : np.ndarray
        2d array (lat, lon) of weights to be used for area weighted means
    ls : dict
        land sea dictionary with keys

        - ["grid_raw"] (2d array (lat, lon) of subsampled land fraction)
        - ["grid_no_ANT"] (grid_raw with Antarctica removed)
    threshold_land : float, default=0.25
        threshold above which land fraction to consider a grid point as a land grid
        point

    Returns
    -------
    var_l : dict
        nested variable at land grid points dictionary with keys

        - [esm] (3d array (run, time, gp_l) of variable at land grid points)
    reg_dict : dict
        region dictionary with added keys

        - ["gps_l"] (2d array (region, gp_l) of region fraction at land grid points)
        - ["wgt_gps_l"] (2d array (region, gp_l) of area weights for each region on land)
        - ["gp_b_l"] (1d array of region index at land grid points with each grid point
          being assigned to a single region)
    ls : dict
        land sea dictionary with added keys

        - ["gp_l"] (1d array of fraction of land at land grid points)
        - ["grid_l"] (2d array (lat, lon) of fraction of land at land grid points)
        - ["idx_grid_l"] (2d boolean array (lat, lon) with land grid points = True for
          plotting on map)
        - ["grid_l_m"] (2d masked array (lat, lon) with ocean masked out for plotting on
          map)
        - ["wgt_gp_l"] (1d array of land area weights, i.e., area weight * land
          fraction)

    """

    # determine which grid points count as land grid points
    idx_l = ls["grid_no_ANT"] > threshold_land

    # extract land points + weights from land sea mask
    ls["gp_l"] = ls["grid_no_ANT"][idx_l]
    ls["grid_l"] = copy.deepcopy(ls["grid_no_ANT"])
    ls["grid_l"][~idx_l] = 0
    ls["idx_grid_l"] = (
        ls["grid_l"] > threshold_land
    )  # gives back a binary (boolean) mask to help with plotting
    ls["grid_l_m"] = np.ma.masked_array(
        ls["grid_l"], mask=np.logical_not(ls["idx_grid_l"])
    )  # masked array (ocean masked out)
    ls["wgt_gp_l"] = (
        wgt[idx_l] * ls["gp_l"]
    )  # weights for land points: multiply weight by land fraction

    # extract the land points + weights from the region grids
    reg_dict["gps_l"] = reg_dict["grids"][:, idx_l]  # country is the first axis
    reg_dict["wgt_gps_l"] = (
        wgt[idx_l] * reg_dict["gps_l"]
    )  # weights for regions (1st axis): region fraction * area weights
    if reg_dict["type"] == "srex" or reg_dict["type"] == "ar6.land":
        reg_dict["wgt_gps_l"] = (
            reg_dict["wgt_gps_l"] * ls["gp_l"]
        )  # * land fraction to account for coastal cells because SREX / ar6.land regions include ocean
    reg_dict["gp_b_l"] = reg_dict["grid_b"][
        idx_l
    ]  # not sure if needed; extracts land from the "binary" mask

    # extract the land points of the variable of interest
    var_l = {}
    for esm in var.keys():
        var_l[esm] = {}
        for scen in var[esm].keys():
            var_l[esm][scen] = var[esm][scen][
                :, :, idx_l
            ]  # run is the first axis, followed by time

    return var_l, reg_dict, ls


def extract_time_period(var, time, start, end):
    """Extract selected time period.

    Parameters
    ----------
    var : np.ndarray
        variable in 1-4d array

        - (time);
        - (run, time);
        - (run, time, gp_l);
        - (run, time, lat, lon)
    time : np.ndarray
        1d array of years
    start : str or int
        first year included in extracted time period
    end : str or int
        last year included in extracted time period

    Returns
    -------
    var_tp : np.ndarray
        variable 1-3d array

        - (time);
        - (time, gp_l);
        - (time, lat, lon);
    time_tp : np.ndarray
        1d array of years of extracted time period

    """

    # find index of start and end of time period
    idx_start = np.where(time == int(start))[0][0]
    idx_end = np.where(time == int(end))[0][0] + 1  # to include the end year

    # extract time period from variable dictionary
    if len(var.shape) > 1:
        var_tp = var[:, idx_start:idx_end]
    else:
        var_tp = var[idx_start:idx_end]

    # extract time period from time vector
    time_tp = time[idx_start:idx_end]

    return var_tp, time_tp



# MESMER, land-climate dynamics group, S.I. Seneviratne
# Copyright (c) 2021 ETH Zurich, MESMER contributors listed in AUTHORS.
# Licensed under the GNU General Public License v3.0 or later see LICENSE or
# https://www.gnu.org/licenses/
"""
Functions to train global trend module of MESMER.
"""


import os

import joblib
import numpy as np
from sklearn.linear_model import LinearRegression
from statsmodels.nonparametric.smoothers_lowess import lowess

from mesmer.io import load_strat_aod


def train_gt(var, targ, esm, time, cfg, save_params=True):
    """
    Derive global trend (emissions + volcanoes) parameters from specified ensemble type
    with specified method.

    Parameters
    ----------
    var : dict
        nested global mean variable dictionary with keys for each scenario employed for
        training

        - [scen] (2d array (run, time) of globally-averaged variable time series)
    targ : str
        target variable (e.g., "tas")
    esm : str
        associated Earth System Model (e.g., "CanESM2" or "CanESM5")
    time : np.ndarray
        [scen] (1d array of years)
    cfg : module
        config file containing metadata
    save_params : bool, default True
        determines if parameters are saved or not, default = True

    Returns
    -------
    params_gt : dict
        dictionary containing the trained parameters for the chosen method / ensemble
        type

        - ["targ"] (emulated variable, str)
        - ["esm"] (Earth System Model, str)
        - ["method"] (applied method, str)
        - ["preds"] (predictors, list of strs)
        - ["scenarios"] (emission scenarios used for training, list of strs)
        - [xx] (additional params depend on method employed)

    Notes
    -----
    - Assumptions:
        - All scens start at the same point in time
        - If historical data is present, historical data and future scenarios are
          transmitted as single time series
    - No perfect smoothness enforced at transition from historical to future scenario
    - No perfect overlap between future scenarios which share the same forcing in the
      beginning is enforced

    """

    # specify necessary variables from config file
    gen = cfg.gen
    method_gt = cfg.methods[targ]["gt"]
    preds_gt = cfg.preds[targ]["gt"]

    scenarios_tr = list(var.keys())

    # initialize parameters dictionary and fill in the metadata which does not depend on the applied method
    params_gt = {}
    params_gt["targ"] = targ
    params_gt["esm"] = esm
    params_gt["method"] = method_gt
    params_gt["preds"] = preds_gt
    params_gt["scenarios"] = scenarios_tr  # single entry in case of ic ensemble

    # apply the chosen method to the type of ensenble
    gt = {}
    if "LOWESS" in params_gt["method"]:
        for scen in scenarios_tr:  # ie derive gt for each scen individually
            gt[scen], frac_lowess_name = train_gt_ic_LOWESS(var[scen])
        params_gt["frac_lowess"] = frac_lowess_name
    else:
        raise ValueError("No alternative method to LOWESS is implemented for now.")

    params_gt["time"] = {}

    if scenarios_tr[0][:2] == "h-":  # ie if hist included

        if gen == 5:
            start_year_fut = 2005
        elif gen == 6:
            start_year_fut = 2014

        idx_start_year_fut = np.where(time[scen] == start_year_fut)[0][0] + 1

        params_gt["time"]["hist"] = time[scen][:idx_start_year_fut]

        # compute median LOWESS estimate of historical part across all scenarios
        gt_lowess_hist_all = np.zeros([len(gt.keys()), len(params_gt["time"]["hist"])])
        for i, scen in enumerate(gt.keys()):
            gt_lowess_hist_all[i] = gt[scen][:idx_start_year_fut]
        gt_lowess_hist = np.median(gt_lowess_hist_all, axis=0)

        if params_gt["method"] == "LOWESS_OLSVOLC":
            scen = scenarios_tr[0]
            var_all = var[scen][:, :idx_start_year_fut]
            for scen in scenarios_tr[1:]:
                var_tmp = var[scen][:, :idx_start_year_fut]
                var_all = np.vstack([var_all, var_tmp])

                # check for duplicates & exclude those runs
                var_all = np.unique(var_all, axis=0)

            params_gt["saod"], params_gt["hist"] = train_gt_ic_OLSVOLC(
                var_all, gt_lowess_hist, params_gt["time"]["hist"], cfg
            )
        elif params_gt["method"] == "LOWESS":
            params_gt["hist"] = gt_lowess_hist

        scenarios_tr_f = list(
            map(lambda x: x.replace("h-", ""), scenarios_tr)
        )  # isolte future scen names

    else:
        idx_start_year_fut = 0  # because first year would be already in future
        scenarios_tr_f = scenarios_tr  # because only future covered anyways

    for scen_f, scen in zip(scenarios_tr_f, scenarios_tr):
        params_gt["time"][scen_f] = time[scen][idx_start_year_fut:]
        params_gt[scen_f] = gt[scen][idx_start_year_fut:]

    # save the global trend paramters if requested
    if save_params:
        dir_mesmer_params = cfg.dir_mesmer_params
        dir_mesmer_params_gt = dir_mesmer_params + "global/global_trend/"
        # check if folder to save params in exists, if not: make it
        if not os.path.exists(dir_mesmer_params_gt):
            os.makedirs(dir_mesmer_params_gt)
            print("created dir:", dir_mesmer_params_gt)
        filename_parts = [
            "params_gt",
            method_gt,
            *preds_gt,
            targ,
            esm,
            *scenarios_tr,
        ]
        filename_params_gt = dir_mesmer_params_gt + "_".join(filename_parts) + ".pkl"
        joblib.dump(params_gt, filename_params_gt)

    return params_gt


def train_gt_ic_LOWESS(var):
    """
    Derive smooth global trend of variable from single ESM ic ensemble with LOWESS
    smoother.

    Parameters
    ----------
    var : np.ndarray
        2d array (run, time) of globally-averaged time series

    Returns
    -------
    gt_lowess : np.ndarray
        1d array of smooth global trend of variable
    frac_lowess : float
        fraction of the data used when estimating each y-value
    """

    # number time steps
    nr_ts = var.shape[1]

    # average across all runs to get a first smoothing
    av_var = np.mean(var, axis=0)

    # apply lowess smoother to further smooth the Tglob time series
    frac_lowess = (
        50 / nr_ts
    )  # rather arbitrarily chosen value that gives a smooth enough trend,
    # open to changes but if much smaller, var trend ends up very wiggly
    frac_lowess_name = "50/nr_ts"

    gt_lowess = lowess(
        av_var, np.arange(nr_ts), return_sorted=False, frac=frac_lowess, it=0
    )

    return gt_lowess, frac_lowess_name


def train_gt_ic_OLSVOLC(var, gt_lowess, time, cfg):
    """
    Derive global trend (emissions + volcanoes) parameters from single ESM ic ensemble
    by adding volcanic spikes to LOWESS trend.

    Parameters
    ----------
    var : np.ndarray
        2d array (run, time) of globally-averaged time series
    gt_lowess : np.ndarray
        1d array of smooth global trend of variable
    time : np.ndarray
        1d array of years
    cfg : module
        config file containing metadata needed to load in stratospheric AOD time series

    Returns
    -------
    coef_saod : float
        stratospheric AOD OLS coefficient for variable variability
    gt : np.ndarray
        1d array of global temperature trend with volcanic spikes

    Notes
    -----
    - Assumptions:
        - only historical time period data is passed

    """

    # specify necessary variables from cfg file
    dir_obs = cfg.dir_obs

    nr_runs, nr_ts = var.shape

    # account for volcanic eruptions in historical time period
    # load in observed stratospheric aerosol optical depth
    aod_obs = load_strat_aod(time, dir_obs).reshape(
        -1, 1
    )  # bring in correct format for sklearn linear regression
    aod_obs_all = np.tile(
        aod_obs, (nr_runs, 1)
    )  # repeat aod time series as many times as runs available
    nr_aod_obs = len(aod_obs)

    if nr_ts != nr_aod_obs:
        raise ValueError(
            f"The number of time steps of the variable ({nr_ts}) and the saod ({nr_aod_obs}) do not match."
        )
    # extract global variability (which still includes volc eruptions) by removing smooth trend from Tglob in historic period
    gv_all_for_aod = np.zeros(nr_runs * nr_aod_obs)
    i = 0
    for run in np.arange(nr_runs):
        gv_all_for_aod[i : i + nr_aod_obs] = var[run] - gt_lowess
        i += nr_aod_obs
    # fit linear regression of gv to aod (because some ESMs react very strongly to volcanoes)
    linreg_gv_volc = LinearRegression(fit_intercept=False).fit(
        aod_obs_all, gv_all_for_aod
    )  # no intercept to not artifically
    # move the ts

    # extract the saod coefficient
    coef_saod = linreg_gv_volc.coef_[0]

    # apply linear regression model to obtain volcanic spikes
    contrib_volc = linreg_gv_volc.predict(aod_obs)

    # merge the lowess trend wit the volc contribution
    gt = gt_lowess + contrib_volc

    return coef_saod, gt



"""
Functions to calibrate all modules of MESMER
"""
import logging
import os
import warnings

from ..create_emulations import create_emus_gt, create_emus_lt, create_emus_lv
from ..io import load_cmipng, load_phi_gc, load_regs_ls_wgt_lon_lat, save_mesmer_bundle
from ..utils import convert_dict_to_arr, extract_land, separate_hist_future
from .train_gt import train_gt
from .train_gv import train_gv
from .train_lt import train_lt
from .train_lv import train_lv

LOGGER = logging.getLogger(__name__)


class _Config:
    """Workaround to mock the ``cfg`` interface used elsewhere"""

    def __init__(
        self,
        esms,
        scenarios,
        cmip_generation,
        cmip_data_root_dir,
        observations_root_dir,
        auxiliary_data_dir,
        reference_period_type,
        reference_period_start_year,
        reference_period_end_year,
        tas_global_trend_method,
        hfds_global_trend_method,
        tas_global_variability_method,
        tas_local_trend_method,
        tas_local_variability_method,
        method_lt_each_gp_sep,
        nr_emus_v,  # TODO: remove when we remove the emulation part
        seeds,  # TODO: remove when we remove the emulation part
        weight_scenarios_equally,
        threshold_land,
        cross_validation_max_iterations,
    ):
        self.esms = esms
        self.scenarios = scenarios
        self.gen = cmip_generation
        # TODO: remove need for trailing seperator eventually
        self.dir_cmipng = f"{cmip_data_root_dir}{os.sep}"
        self.dir_obs = f"{observations_root_dir}{os.sep}"
        self.dir_aux = f"{auxiliary_data_dir}{os.sep}"
        self.ref = {
            "type": reference_period_type,
            "start": reference_period_start_year,
            "end": reference_period_end_year,
        }
        self.methods = {
            "tas": {
                "gt": tas_global_trend_method,
                "gv": tas_global_variability_method,
                "lt": tas_local_trend_method,
                "lv": tas_local_variability_method,
            },
            "hfds": {"gt": hfds_global_trend_method},
        }
        self.method_lt_each_gp_sep = method_lt_each_gp_sep

        # Essentially metadata about predictors used. Maybe used for naming etc.
        # TODO: work out where and how this is used in future so it can be defined
        # more precisely.
        self.preds = {
            "tas": {
                "gt": ["saod"],
                "gv": [],
            },
            "hfds": {"gt": []},
        }
        self.preds["tas"]["g_all"] = self.preds["tas"]["gt"] + self.preds["tas"]["gv"]

        self.nr_emus_v = nr_emus_v
        self.seed = seeds
        self.wgt_scen_tr_eq = weight_scenarios_equally
        self.threshold_land = threshold_land
        self.max_iter_cv = cross_validation_max_iterations


# TODO: remove draw realisations functionality
def _calibrate_and_draw_realisations(
    esms,
    scenarios_to_train,
    target_variable,
    reg_type,
    threshold_land,
    output_file,
    scen_seed_offset_v,  # TODO: remove when we remove the emulation part
    cmip_data_root_dir,
    observations_root_dir,
    auxiliary_data_dir,
    cmip_generation=6,
    reference_period_type="individ",
    reference_period_start_year="1850",
    reference_period_end_year="1900",
    tas_global_trend_method="LOWESS_OLSVOLC",
    hfds_global_trend_method="LOWESS",
    tas_global_variability_method="AR",
    tas_local_trend_method="OLS",
    tas_local_variability_method="OLS_AR1_sci",
    # specify if the local trends method is applied to each grid point separately.
    # Currently it must be set to True
    method_lt_each_gp_sep=True,
    nr_emus_v=100,  # TODO: remove when we remove the emulation part
    weight_scenarios_equally=True,
    cross_validation_max_iterations=30,
):
    """calibrate mesmer - additional predictors configuration. used for end-to-end test"""

    tas_g_dict = {}  # tas with global coverage
    GSAT_dict = {}  # global mean tas
    GHFDS_dict = {}  # global mean hfds (needed as predictor)
    tas_g = {}
    GSAT = {}
    GHFDS = {}
    time = {}

    # TODO: decide if we want this functionality and, if we do, test what
    #       happens if scen_seed_offset_v != 0
    if scen_seed_offset_v == 0:
        scenarios_emus_v = ["all"]
    else:
        scenarios_emus_v = scenarios_to_train

    seeds = {}
    i = 0
    for esm in esms:
        seeds[esm] = {}
        j = 0
        for scen in scenarios_emus_v:
            seeds[esm][scen] = {}
            seeds[esm][scen]["gv"] = i + j * scen_seed_offset_v
            seeds[esm][scen]["lv"] = i + j * scen_seed_offset_v + 1000000
            j += 1
        i += 1

    cfg = _Config(
        esms,
        scenarios_to_train,
        cmip_generation,
        cmip_data_root_dir,
        observations_root_dir,
        auxiliary_data_dir,
        reference_period_type,
        reference_period_start_year,
        reference_period_end_year,
        tas_global_trend_method,
        hfds_global_trend_method,
        tas_global_variability_method,
        tas_local_trend_method,
        tas_local_variability_method,
        method_lt_each_gp_sep,
        nr_emus_v,
        seeds,
        weight_scenarios_equally,
        threshold_land,
        cross_validation_max_iterations,
    )

    for esm in esms:
        LOGGER.info("Loading data for %s", esm)
        tas_g_dict[esm] = {}
        GSAT_dict[esm] = {}
        GHFDS_dict[esm] = {}
        time[esm] = {}

        for scen in scenarios_to_train:
            # TODO: rename tas_g_tmp to target_variable_g_tmp or simply
            #       hard-code tas as always being the target variable
            tas_g_tmp, GSAT_tmp, lon_tmp, lat_tmp, time_tmp = load_cmipng(
                target_variable, esm, scen, cfg
            )

            if tas_g_tmp is None:
                # should this be an error?
                warnings.warn(f"Scenario {scen} does not exist for tas for ESM {esm}")
            else:  # if scen exists: save fields + load hfds fields for it too
                (
                    tas_g_dict[esm][scen],
                    GSAT_dict[esm][scen],
                    lon,
                    lat,
                    time[esm][scen],
                ) = (
                    tas_g_tmp,
                    GSAT_tmp,
                    lon_tmp,
                    lat_tmp,
                    time_tmp,
                )
                _, GHFDS_dict[esm][scen], _, _, _ = load_cmipng("hfds", esm, scen, cfg)

        tas_g[esm] = convert_dict_to_arr(tas_g_dict[esm])
        GSAT[esm] = convert_dict_to_arr(GSAT_dict[esm])
        GHFDS[esm] = convert_dict_to_arr(GHFDS_dict[esm])

    # load in the constant files
    reg_dict, ls, wgt_g, lon, lat = load_regs_ls_wgt_lon_lat(reg_type, lon, lat)

    # extract land
    tas, reg_dict, ls = extract_land(
        tas_g, reg_dict, wgt_g, ls, threshold_land=threshold_land
    )

    for esm in esms:
        LOGGER.info("Calibrating %s", esm)

        LOGGER.info("Calibrating global trend module")
        # TODO: `target_variable` is used here but not elsewhere (where tas is
        #       basically hard-coded)
        params_gt_T = train_gt(
            GSAT[esm], target_variable, esm, time[esm], cfg, save_params=False
        )
        # TODO: remove hard-coded hfds
        params_gt_hfds = train_gt(
            GHFDS[esm], "hfds", esm, time[esm], cfg, save_params=False
        )

        # From params_gt_T, extract the global-trend so that the global
        # variability, local trends, and local variability modules can be
        # trained.
        # In this case we're not actually creating emulations
        LOGGER.info("Creating global-trend emulations")
        preds_gt = {"time": time[esm]}

        # TODO: remove use of emus_gt from this script.
        emus_gt_T = create_emus_gt(
            params_gt_T, preds_gt, cfg, concat_h_f=True, save_emus=False
        )
        gt_T_s = create_emus_gt(
            params_gt_T, preds_gt, cfg, concat_h_f=False, save_emus=False
        )

        LOGGER.info(
            "Preparing predictors for global variability, local trends, and "
            "local variability"
        )
        gt_T2_s = {}
        for scen in gt_T_s.keys():
            gt_T2_s[scen] = gt_T_s[scen] ** 2

        gt_hfds_s = create_emus_gt(
            params_gt_hfds, preds_gt, cfg, concat_h_f=False, save_emus=False
        )

        gv_novolc_T = {}
        for scen in emus_gt_T.keys():
            gv_novolc_T[scen] = GSAT[esm][scen] - emus_gt_T[scen]

        gv_novolc_T_s, time_s = separate_hist_future(gv_novolc_T, time[esm], cfg)

        tas_s, time_s = separate_hist_future(tas[esm], time[esm], cfg)

        LOGGER.info("Calibrating global variability module")
        params_gv_T = train_gv(
            gv_novolc_T_s, target_variable, esm, cfg, save_params=False
        )

        # TODO: remove because time_v is not needed for calibration
        time_v = {}
        time_v["all"] = time[esm][scen]

        LOGGER.info("Calibrating local trends module")
        preds = {
            "gttas": gt_T_s,
            "gttas2": gt_T2_s,
            "gthfds": gt_hfds_s,
            "gvtas": gv_novolc_T_s,
        }
        targs = {"tas": tas_s}
        params_lt, params_lv = train_lt(preds, targs, esm, cfg, save_params=False)

        # Create forced local warming samples used for training the local variability
        # module. Samples are cheap to create so not an issue to have here.
        LOGGER.info("Creating local trends emulations")
        preds_lt = {"gttas": gt_T_s, "gttas2": gt_T2_s, "gthfds": gt_hfds_s}
        lt_s = create_emus_lt(
            params_lt, preds_lt, cfg, concat_h_f=False, save_emus=False
        )

        LOGGER.info("Calibrating local variability module")
        # derive variability part induced by gv
        preds_lv = {"gvtas": gv_novolc_T_s}  # predictors_list

        # Create local variability due to global variability warming samples
        # used for training the local variability module. Samples are cheap to create so not an issue to have here.
        lv_gv_s = create_emus_lv(
            params_lv, preds_lv, cfg, save_emus=False, submethod="OLS"
        )

        # tas essentially hard-coded here too
        LOGGER.debug(
            "Calculating residual variability i.e. what remains of tas once "
            "lt + gv removed"
        )
        res_lv_s = {}  # residual local variability
        for scen in tas_s.keys():
            res_lv_s[scen] = tas_s[scen] - lt_s[scen]["tas"] - lv_gv_s[scen]["tas"]

        LOGGER.debug("Loading auxiliary files")
        aux = {}
        aux["phi_gc"] = load_phi_gc(
            lon, lat, ls, cfg, L_start=1750, L_end=2000, L_interval=250
        )  # better results with default values L, but like this much faster + less space needed

        LOGGER.debug("Finalising training of local variability module on derived data")
        targs_res_lv = {"tas": res_lv_s}
        params_lv = train_lv(
            {}, targs_res_lv, esm, cfg, save_params=False, aux=aux, params_lv=params_lv
        )

        save_mesmer_bundle(
            output_file,
            params_lt,
            params_lv,
            params_gv_T,
            land_fractions=ls["grid_l_m"],
            lat=lat["c"],
            lon=lon["c"],
        )



# MESMER, land-climate dynamics group, S.I. Seneviratne
# Copyright (c) 2021 ETH Zurich, MESMER contributors listed in AUTHORS.
# Licensed under the GNU General Public License v3.0 or later see LICENSE or
# https://www.gnu.org/licenses/
"""
Functions to load in cmip5 and cmip6 data from the cmip-ng archive at ETHZ.
"""

import copy as copy
import glob as glob

import numpy as np
import xarray as xr


def extract_time_lon_lat_wgt3d(data):
    """
    Extract time, longitude, latitude, and 3d weights from file from ETHZ cmip-ng
    archive.

    Parameters
    ----------
    data : xr.core.dataset.Dataset
        dataset to extract time, lon, lat, and wgt3d from

    Returns
    -------
    time : np.ndarray
        1d array of years
    lon : dict
        longitude dictionary with key
     - ["c"] (1d array with longitudes at center of grid cell)
    lat : dict
        latitude dictionary with key

        - ["c"] (1d array with latitudes at center of grid cell)
    wgt3d : np.ndarray
        3d array (time, lat, lon) of area weight of each grid point
    """

    # extract time
    time = data.time.dt.year.values

    # extract longitude and latitude at center of grid cell vectors
    lon = {}
    lat = {}
    lon["c"] = data.lon.values  # longitude at center (c) of grid cell
    lat["c"] = data.lat.values

    lons, lats = np.meshgrid(lon["c"], lat["c"])
    wgt2d = np.cos(np.deg2rad(lats))
    wgt3d = np.tile(wgt2d, [len(time), 1]).reshape(
        [len(time), wgt2d.shape[0], wgt2d.shape[1]]
    )

    return time, lon, lat, wgt3d


def find_files_cmipng(gen, esm, var, scenario, dir_cmipng):
    """Find filname in ETHZ cmip-ng archive.

    Parameters
    ----------
    gen : int
        generation (5 or 6)
    esm : str
        Earth System Model (e.g., "CanESM2" or "CanESM5")
    var : str
        variable (e.g., "tas", "tos")
    scenario : str
        scenario (e.g., "rcp85" or "ssp585")
    dir_cmipng : str
        path to cmip-ng archive

    Returns
    -------
    path_runs_list : list
        list of paths to all the filenames found for the given input

    Notes
    -----
    - Not fool-proof enough yet for ESMs with different forcings. Could/ should be
      improved if there is time.
    - TODO:
        - improve and extend list of excluded runs/ ESMs for all cmip generations and
          variables
    """

    # for cmip5-ng
    if gen == 5:
        dir_name = dir_cmipng + var + "/"

        if var == "tas":
            esms_excl = ["GISS-E2-H"]  # list of all ESMs which have excluded runs
            runs_excl = ["tas_ann_GISS-E2-H_rcp85_r2i1p1_g025"]  # list of excluded runs
        else:
            esms_excl = []
            runs_excl = []
            print("TO DO: create list of excluded runs / ESMs for this variable")

        path_runs_list = sorted(
            glob.glob(
                dir_name
                + var
                + "_ann_"
                + esm
                + "_"
                + scenario
                + "_"
                + "r*i1p1"
                + "_g025.nc"
            )
        )

    # for cmip6-ng
    if gen == 6:
        dir_name = dir_cmipng + var + "/ann/g025/"

        # TODO: remove hard-coding
        if var == "tas":
            esms_excl = [
                "ACCESS-ESM1-5",
                "IPSL-CM6A-LR",
                "UKESM1-0-LL",
            ]  # list of all ESMs which have excluded runs
            runs_excl = [
                "tas_ann_ACCESS-ESM1-5_ssp245_r21i1p1f1_g025.nc",
                "tas_ann_IPSL-CM6A-LR_ssp434_r1i1p1f2_g025.nc",
                "tas_ann_IPSL-CM6A-LR_ssp460_r1i1p1f2_g025.nc",
                "tas_ann_UKESM1-0-LL_ssp126_r5i1p1f2_g025.nc",
                "tas_ann_UKESM1-0-LL_ssp126_r6i1p1f2_g025.nc",
                "tas_ann_UKESM1-0-LL_ssp126_r7i1p1f2_g025.nc",
                "tas_ann_UKESM1-0-LL_ssp370_r5i1p1f2_g025.nc",
                "tas_ann_UKESM1-0-LL_ssp370_r6i1p1f2_g025.nc",
                "tas_ann_UKESM1-0-LL_ssp370_r7i1p1f2_g025.nc",
            ]  # list of excluded runs
            # TODO: extend list of excluded runs / ESMs
        elif var == "hfds":
            esms_excl = ["ACCESS-ESM1-5", "IPSL-CM6A-LR", "UKESM1-0-LL"]
            runs_excl = [
                "hfds_ann_ACCESS-ESM1-5_ssp126_r7i1p1f1_g025.nc",
                "hfds_ann_ACCESS-ESM1-5_ssp245_r7i1p1f1_g025.nc",
                "hfds_ann_ACCESS-ESM1-5_ssp370_r7i1p1f1_g025.nc",
                "hfds_ann_ACCESS-ESM1-5_ssp585_r7i1p1f1_g025.nc",
                "hfds_ann_IPSL-CM6A-LR_ssp434_r1i1p1f2_g025.nc",
                "hfds_ann_IPSL-CM6A-LR_ssp460_r1i1p1f2_g025.nc",
                "hfds_ann_UKESM1-0-LL_ssp126_r5i1p1f2_g025.nc",
                "hfds_ann_UKESM1-0-LL_ssp126_r6i1p1f2_g025.nc",
                "hfds_ann_UKESM1-0-LL_ssp126_r7i1p1f2_g025.nc",
                "hfds_ann_UKESM1-0-LL_ssp370_r5i1p1f2_g025.nc",
                "hfds_ann_UKESM1-0-LL_ssp370_r6i1p1f2_g025.nc",
                "hfds_ann_UKESM1-0-LL_ssp370_r7i1p1f2_g025.nc",
            ]
        else:
            esms_excl = []
            runs_excl = []
            print("TO DO: create list of excluded runs / ESMs for this variable")

        # print(
        #   "TO DO: rewrite selection of p2 in way that needs less space for CanESM5 (+ see if other ESMs need p2 too)"
        # )
        path_runs_list_scen = sorted(
            glob.glob(
                dir_name
                + var
                + "_ann_"
                + esm
                + "_"
                + scenario
                + "_"
                + "r*i1p1f*"
                + "_g025.nc"
            )
        )

        path_runs_list_hist = sorted(
            glob.glob(
                dir_name
                + var
                + "_ann_"
                + esm
                + "_historical_"
                + "r*i1p1f*"
                + "_g025.nc"
            )
        )

        # check if both scenario and historical run are available for this realization, if yes add to path_runs_list
        path_runs_list = []
        for run in path_runs_list_scen:
            run_hist = run.replace(scenario, "historical")
            if run_hist in path_runs_list_hist:
                path_runs_list.append(run)

        # TODO: redecide if I am fine with CanESM5 p2 but not all scenarios or if I prefer the worse p1 which has all scenarios
        # code below = old version when used p2 instead
        # if esm != "CanESM5":
        #    path_runs_list = sorted(
        #       glob.glob(
        #   dir_name
        #  + var
        # + "_ann_"
        # + esm
        # + "_"
        #      + scenario
        #     + "_"
        #    + "r*i1p1f*"
        #   + "_g025.nc"
        #   )
        # )

        # else:
        #   path_runs_list = sorted(
        #      glob.glob(
        #         dir_name
        #        + var
        #       + "_ann_"
        #      + esm
        #     + "_"
        #    + scenario
        #   + "_"
        #  + "r*i1p2f*"
        # + "_g025.nc"
        #    )
        # )
    if len(path_runs_list) == 0:  # if no entries found, return the empty list
        return path_runs_list

    # ordering does not work for ESMs with > 9 runs -> find first run + put in front
    index_first = [i for i, s in enumerate(path_runs_list) if "r1i1" in s][
        0
    ]  # find first run
    path_runs_list.insert(
        0, path_runs_list.pop(index_first)
    )  # move first run to begin of list

    # exclude faulty runs in archive
    if esm in esms_excl:
        for run_excl in runs_excl:
            for run in path_runs_list:
                if run_excl in run:
                    path_runs_list.remove(run)

    return path_runs_list


def load_cmipng(targ, esm, scen, cfg):
    """Load ESM runs from cmip-ng archive at ETHZ.

    Parameters
    ----------
    targ : str
        target variable (e.g., "tas")
    esm : str
        Earth System Model (e.g., "CanESM2" or "CanESM5")
    scen : str
        future scenario (e.g., "rcp85" or "ssp585")
    cfg : module
        config file containing metadata

    Returns
    -------
    targ : dict
        target variable anomaly dictionary with keys

        - [run] (3d array (time, lat, lon) of variable)
    GTARG : dict
        area-weighted global mean target variable anomaly dictionary with keys

        - [run]  (1d array of globally-averaged variable anomaly time series)
    lon : dict
        longitude dictionary with key

        - ["c"] (1d array with longitudes at center of grid cell)
    lat : dict
        latitude dictionary with key

        - ["c"] (1d array with latitudes at center of grid cell)
    time : np.ndarray
        1d array of years
    """

    # check if scenario describes
    if scen[:2] == "h-":
        scen_fut = scen[2:]
    else:
        raise ValueError(
            "No version without historical time period is currently implemented."
        )

    targ_func_mapping = {"hfds": load_cmipng_hfds, "tas": load_cmipng_tas}
    # once start working with other vars, extend dict eg {"pr": load_cmipng_pr, "hfds": load_cmipng_hfds, "tas": load_cmipng_tas}

    load_targ = targ_func_mapping[targ]

    targ, GTARG, lon, lat, time = load_targ(esm, scen_fut, cfg)

    return targ, GTARG, lon, lat, time


def load_cmipng_file(run_path, gen, scen):
    """Load file in ETHZ cmip-ng archive.

    Parameters
    ----------
    run_path : str
        path to file
    gen : int
        generation (5 or 6)
    scen : str
        future scenario (e.g., "rcp85" or "ssp585")

    Returns
    -------
    data : xr.core.dataset.Dataset
        loaded dataset
    run : int
        realization index
    """

    # account for difference in naming convention in cmipx-ng archives
    if gen == 5:
        data = (
            xr.open_dataset(run_path)
            .rename({"year": "time"})
            .roll(lon=72, roll_coords=True)
        )  # rename to time for consistency with cmip6, roll so land in center
        data = data.assign_coords(
            lon=(((data.lon + 180) % 360) - 180)
        )  # assign_coords so that labels = reasonable
        run = int(
            data.attrs["source_ensemble"].split("r")[1].split("i")[0]
        )  # extract ens member

    if gen == 6:
        if "ssp534-over" in run_path:
            run_path_ssp_534over = run_path
            run_path_ssp_585 = run_path.replace(scen, "ssp585")
            run_path_hist = run_path.replace(scen, "historical")
            data = xr.open_mfdataset(
                [run_path_hist, run_path_ssp_585, run_path_ssp_534over],
                combine="by_coords",
                concat_dim="time",
                preprocess=preprocess_ssp534over,
            )
        else:  # for every other scenario
            run_path_ssp = run_path
            run_path_hist = run_path.replace(scen, "historical")
            data = xr.open_mfdataset(
                [run_path_hist, run_path_ssp], combine="by_coords", concat_dim="time"
            )

        data = data.roll(lon=72, roll_coords=True)  # roll so land in center
        data = data.assign_coords(
            lon=(((data.lon + 180) % 360) - 180)
        )  # assign_coords so that labels = reasonable
        data = data.sortby(["lat", "lon"])
        run = data.attrs["realization_index"]

    return data, run


def load_cmipng_hfds(esm, scen, cfg):
    """Load ESM hfds runs from cmip-ng archive at ETHZ.

    Parameters
    ----------
    esm : str
        Earth System Model (e.g., "CanESM2" or "CanESM5")
    scen : str
        future scenario (e.g., "rcp85" or "ssp585")
    cfg : module
        config file containnig metadata

    Returns
    -------
    hfds : dict
        Downward Heat Flux at Sea Water Surface (hfds) anomaly dictionary with keys

        - [run] (3d array (time, lat, lon) of variable)
    GHFDS : dict
        area-weighted global mean hfds anomaly dictionary with keys

        - [run]  (1d array of globally-averaged tas anomaly time series)
    lon : dict
        longitude dictionary with key

        - ["c"] (1d array with longitudes at center of grid cell)
    lat : dict
        latitude dictionary with key

        - ["c"] (1d array with latitudes at center of grid cell)
    time : np.ndarray
        1d array of years

    Notes
    -----
    - There are some overlaps with load_cmipng_tas()
    - These functions could be more optimally merged in the future to avoid repetition

    """

    # specify necessary variables from cfg file
    gen = cfg.gen
    ref = cfg.ref
    dir_cmipng = cfg.dir_cmipng

    # find the files which fulfill the specifications
    path_runs_list = find_files_cmipng(gen, esm, "hfds", scen, dir_cmipng)

    # exit function in case path_runs_list is empty (ie no file matches the search criterion)
    if len(path_runs_list) == 0:
        hfds = GHFDS = lon = lat = time = None
        return hfds, GHFDS, lon, lat, time

    # load hfds on global grid and compute anomalies thereof
    hfds = {}
    run_nrs = {}

    for run_path in path_runs_list:
        # account for difference in naming convention in cmipx-ng archives
        data, run = load_cmipng_file(run_path, gen, scen)

        run_nrs[run_path] = run  # tmp saved this way for deriving anomalies later
        hfds[run] = copy.deepcopy(data.hfds.values)

        if ref["type"] == "all":
            if run_path == path_runs_list[0]:  # for the first run need to initialize
                hfds_ref = (
                    data.hfds.sel(time=slice(ref["start"], ref["end"]))
                    .mean(dim="time")
                    .values
                    * 1.0
                    / len(path_runs_list)
                )
            else:
                hfds_ref += (
                    data.hfds.sel(time=slice(ref["start"], ref["end"]))
                    .mean(dim="time")
                    .values
                    * 1.0
                    / len(path_runs_list)
                )  # sum up all ref climates

        if ref["type"] == "individ":
            hfds_ref = (
                data.hfds.sel(time=slice(ref["start"], ref["end"]))
                .mean(dim="time")
                .values
            )
            hfds[run] -= hfds_ref  # compute anomalies

        if ref["type"] == "first" and run == "1":
            # TO CHECK
            hfds_ref = (
                data.hfds.sel(time=slice(ref["start"], ref["end"]))
                .mean(dim="time")
                .values
            )

    if ref["type"] == "all" or ref["type"] == "first":
        for run_path in path_runs_list:
            run = run_nrs[run_path]
            hfds[run] -= hfds_ref  # compute anomalies

    # extract time, longitude, latitude, and 3d weights from data
    time, lon, lat, wgt3d = extract_time_lon_lat_wgt3d(data)

    # compute area-weighted GHFDS from global hfds
    # ATTENTION: currently not implemented to account for land fraction within grid cells.
    # I.e., coastal grid cells count as full ocean grid cells here. Expected to have negligible impact on global mean.
    GHFDS = {}
    for run_path in path_runs_list:
        run = run_nrs[run_path]
        masked_hfds = np.ma.masked_array(
            hfds[run], np.isnan(hfds[run])
        )  # account for missing values over land
        GHFDS[run] = np.average(
            masked_hfds, axis=(1, 2), weights=wgt3d
        ).data  # no need to keep mask since False everywhere

    return hfds, GHFDS, lon, lat, time


def load_cmipng_tas(esm, scen, cfg):
    """Load ESM tas runs from cmip-ng archive at ETHZ.

    Parameters
    ----------
    esm : str
        Earth System Model (e.g., "CanESM2" or "CanESM5")
    scen : str
        future scenario (e.g., "rcp85" or "ssp585")
    cfg : module
        config file containing metadata

    Returns
    -------
    tas : dict
        2-m air temperature anomaly dictionary with keys

        - [run] (3d array (time, lat, lon) of variable)
    GSAT : dict
        area-weighted global mean 2-m air temperature anomaly dictionary with keys

        - [run]  (1d array of globally-averaged tas anomaly time series)
    lon : dict
        longitude dictionary with key

        - ["c"] (1d array with longitudes at center of grid cell)
    lat : dict
        latitude dictionary with key

        - ["c"] (1d array with latitudes at center of grid cell)
    time : np.ndarray
        1d array of years

    Notes
    -----
    - There are some overlaps with load_cmipng_hfds()
    - These functions could be more optimally merged in the future to avoid repetition

    """

    # specify necessary variables from cfg file
    gen = cfg.gen
    ref = cfg.ref
    dir_cmipng = cfg.dir_cmipng

    # find the files which fulfill the specifications
    path_runs_list = find_files_cmipng(gen, esm, "tas", scen, dir_cmipng)

    # exit function in case path_runs_list is empty (ie no file matches the search criterion)
    if len(path_runs_list) == 0:
        tas = GSAT = lon = lat = time = None
        return tas, GSAT, lon, lat, time

    # load 2-m air temperature on global grid and compute anomalies thereof
    tas = {}
    run_nrs = {}

    for run_path in path_runs_list:
        # account for difference in naming convention in cmipx-ng archives
        data, run = load_cmipng_file(run_path, gen, scen)

        run_nrs[run_path] = run  # tmp saved this way for deriving anomalies later
        tas[run] = copy.deepcopy(data.tas.values)

        if ref["type"] == "all":
            if run_path == path_runs_list[0]:  # for the first run need to initialize
                tas_ref = (
                    data.tas.sel(time=slice(ref["start"], ref["end"]))
                    .mean(dim="time")
                    .values
                    * 1.0
                    / len(path_runs_list)
                )
            else:
                tas_ref += (
                    data.tas.sel(time=slice(ref["start"], ref["end"]))
                    .mean(dim="time")
                    .values
                    * 1.0
                    / len(path_runs_list)
                )  # sum up all ref climates

        if ref["type"] == "individ":
            tas_ref = (
                data.tas.sel(time=slice(ref["start"], ref["end"]))
                .mean(dim="time")
                .values
            )
            tas[run] -= tas_ref  # compute anomalies

        if ref["type"] == "first" and run == "1":
            # TO CHECK
            tas_ref = (
                data.tas.sel(time=slice(ref["start"], ref["end"]))
                .mean(dim="time")
                .values
            )

    if ref["type"] == "all" or ref["type"] == "first":
        for run_path in path_runs_list:
            run = run_nrs[run_path]
            tas[run] -= tas_ref  # compute anomalies

    # extract time, longitude, latitude, and 3d weights from data
    time, lon, lat, wgt3d = extract_time_lon_lat_wgt3d(data)

    # compute area-weighted GSAT from global tas
    GSAT = {}
    for run_path in path_runs_list:
        run = run_nrs[run_path]
        GSAT[run] = np.average(tas[run], axis=(1, 2), weights=wgt3d)

    return tas, GSAT, lon, lat, time


def preprocess_ssp534over(ds):
    """
    Preprocess datasets to manage to combine historical, ssp585, and ssp534-over into
    single time series.

    Parameters
    ----------
    ds : xr.Dataset
        dataset to be concatenated with other datasets

    Returns
    -------
    ds : xr.Dataset
        dataset cut after 2039 unless the start year is after 2030

    Notes
    -----
    - ssp534over starts in 2040, before it follows ssp585
    - This pre-processing allows to concatenate the individual files without running
      into overlapping time periods
    - Code received from Mathias Hauser on 20201117 (personal exchange via slack)

    """

    first_year = ds.time.dt.year[0]

    if first_year < 2030:
        ds = ds.sel(time=slice(None, "2039"))

    return ds
